# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Creates a chat game base class as an OpenSpiel Environment."""

import collections
import dataclasses
import string

from typing import Any, Callable, Dict, OrderedDict, List, Tuple, Union
from absl import logging
import numpy as np

from open_spiel.python.games.chat_games.envs.observations import utils as observation_utils
from open_spiel.python.games.chat_games.envs.payoffs import utils as payoff_utils
from open_spiel.python.games.chat_games.envs.termination import utils as term_utils
from open_spiel.python.games.chat_games.envs.utils import header as header_utils
from open_spiel.python.games.chat_games.envs.utils import text

from open_spiel.python.games.chat_games.utils import logging_utils

import pyspiel


ct = logging_utils.ColorText()

REWARD_MODEL = pyspiel.GameType.RewardModel.TERMINAL

ALL_PLAYERS = 'Everyone'

MIN_RND_SEED = 42
MAX_RND_SEED = 9999
DEFAULT_LLM_SEED = 42

LLM_LENGTH_MESSAGE_TOKENS = 300
LLM_LENGTH_MESSAGE_CHARS = 300
LLM_LENGTH_OBS_TOKENS = 300
LLM_LENGTH_OBS_CHARS = 300
LLM_LENGTH_PAYOFF_OBS_TOKENS = 300
LLM_LENGTH_PAYOFF_OBS_CHARS = 300

LLM_LENGTH_LIST_OF_WORDS_TOKENS = 30
LLM_LIST_GEN_ATTEMPTS = 30

LLM_LENGTH_SCORE_TOKENS = 10

ITEM_PREFIX = '* '

MIN_PLAYERS = 2  # any less and it's not a game, is it ;)
MAX_PLAYERS = 10  # this is set arbitrarily for now, should be upper bound
MAX_NUM_REPLIES = 5

VEC_SIZE = 100  # int, length of vector returned by `vectorize` on string input

DEFAULT_PARAMS = {'num_distinct_actions': 2,
                  'num_llm_seeds': 1,
                  'num_init_states': 1,
                  'num_players': MIN_PLAYERS,
                  'players': 0,  # open_spiel tests use this for `num_players`
                  'min_utility': -10.0,
                  'max_utility': 10.0,
                  'num_max_replies': 1,
                  'silence_logging': True}

GAME_TYPE_KWARGS = {
    'dynamics': pyspiel.GameType.Dynamics.SEQUENTIAL,
    'chance_mode': pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC,
    'information': pyspiel.GameType.Information.IMPERFECT_INFORMATION,
    'reward_model': REWARD_MODEL,
    'max_num_players': MAX_PLAYERS,
    'min_num_players': MIN_PLAYERS,
    'provides_observation_string': True,
    'provides_observation_tensor': True,
    'provides_factored_observation_string': True,
    'parameter_specification': DEFAULT_PARAMS,
    'default_loadable': True
    }

GAME_TYPE = pyspiel.GameType(
        short_name='chat_game',
        long_name='Chat Game',
        utility=pyspiel.GameType.Utility.GENERAL_SUM,
        provides_information_state_string=False,
        provides_information_state_tensor=False,
        **GAME_TYPE_KWARGS)


@dataclasses.dataclass(frozen=True)
class InitialStateConfiguration:
  """Constructor.

  Attributes:
    actions: dict, {'player_names': list of str,
                    <prompt_action_i>: list of str,
                    ...}
    seeds: list of ints, llm seeds (chance nodes)
    scenario_prompt: str, initial message with header (no tone)
    private_info: dict mapping info-type to list of str, one for each player
      i.e., private (prior) info available to each player
  """
  actions: OrderedDict[str, List[str]]
  seeds: List[int]
  scenario_prompt: str
  private_info: OrderedDict[str, List[str]]


class ChatGameState(pyspiel.State):
  """Chat game state."""

  def __init__(self,
               game: ...,
               init_state_configs: Tuple[InitialStateConfiguration, ...]):
    """Constructor.

    Args:
      game: see ChatGame class (should inherit from BaseChatGame)
      init_state_configs: tuple of InitialStateConfiguration objects
        one will be selected at random to initialize the game
    """
    super().__init__(game)  # access game with self.get_game()

    self._init_state_configs = init_state_configs

    # Init empty game w/ init_state_configs[0]. Overwrite later in game setup.
    self._init_empty_game(init_state_configs[0])
    self._game_setup = False

    self._llm_termination = False

    self._rnd = self.get_game().rnd

    self._played_actions = []
    self._current_speaker = 1
    self._current_player = 1
    self._speakers = []
    self._num_actions_played = 0
    self._returns = None
    self._player_action = None

  def _init_empty_game(self, init_state_config: InitialStateConfiguration):
    """Initialize an empty game.

    Args:
      init_state_config: InitialStateConfiguration object
    """
    actions = init_state_config.actions
    seeds = init_state_config.seeds
    private_info = init_state_config.private_info

    self._num_actions = tuple([0 for _ in actions.values()])
    prompt_action_vals = [
        '' for _ in self.get_game().header.action_keys
    ]
    self._prompt_actions = OrderedDict(zip(self.get_game().header.action_keys,
                                           prompt_action_vals))
    self._names = ['' for _ in actions['player_names']]

    self._llm_seeds = [0 for _ in seeds]
    assert self.get_game().num_llm_seeds == len(self._llm_seeds)

    self._scenario_prompt = ''

    empty_pi_vals = []
    for pi in private_info.values():
      empty_player_pi = ['' for _ in pi]
      empty_pi_vals.append(empty_player_pi)
    self._private_info = OrderedDict(zip(private_info.keys(), empty_pi_vals))

    self._dialogue = ['']

  def _setup_game(self, init_state_config: InitialStateConfiguration):
    """Set up the game.

    Args:
      init_state_config: InitialStateConfiguration object
    """
    actions = init_state_config.actions
    seeds = init_state_config.seeds
    scenario_prompt = init_state_config.scenario_prompt
    private_info = init_state_config.private_info

    self._num_actions = tuple([len(a) for a in actions.values()])
    prompt_action_vals = [
        actions[key] for key in self.get_game().header.action_keys
    ]
    self._prompt_actions = OrderedDict(zip(self.get_game().header.action_keys,
                                           prompt_action_vals))
    self._names = actions['player_names']

    self._llm_seeds = seeds
    assert self.get_game().num_llm_seeds == len(self._llm_seeds)

    self._scenario_prompt = scenario_prompt

    self._private_info = private_info

    self._dialogue = [scenario_prompt]

  def __str__(self):
    """String for debug purposes. No particular semantics are required."""
    if not self._game_setup:
      return 'Setting up game...'
    else:
      return self._dialogue[-1]

  def _unravel_flat_action(self, action: int) -> Tuple[int, ...]:
    """Returns an action tuple with action types separated.

    Args:
      action: int
    Returns:
      action_tuple: tuple of ints, each int represents a separate component of
        the combinatorial action-space  
    """
    idxs = np.unravel_index([action], self._num_actions)
    return tuple([idx[0] for idx in idxs])

  def _build_payoff_query(self,
                          payoff_query: str,
                          msg: str,
                          player_str: str) -> str:
    """Construct prompt for LLM to perform sentiment analysis.

    Args:
      payoff_query: str, query to be formatted for llm
      msg: str, message to be analyzed
      player_str: str, player message is analyzed (scored) for
    Returns:
      str: str, payoff prompt to feed to LLM
    """
    payoff_dict = {'m': msg, 'p': player_str}
    return payoff_query.format(**payoff_dict)

  def _llm_is_terminal(self) -> bool:
    ct.set_color(logging_utils.RED)
    prefix = self.get_game().llm_termination_prompt.obs_trans_prefix
    postfix = self.get_game().llm_termination_prompt.obs_trans_postfix
    if prefix or postfix:
      prompt = prefix + self.dialogue_str + postfix
      term_obs = self.get_game().generate_response(prompt,
                                                   seed=DEFAULT_LLM_SEED)
      logging.info(ct.color('LLM summary:\n%s'), term_obs)
    else:
      term_obs = self.dialogue_str
    llm_termination = self.get_game().generate_bool(
        self.get_game().llm_termination_prompt.query.format(msg=term_obs),
        seed=DEFAULT_LLM_SEED)
    logging.info(ct.color('LLM termination condition met? %s'),
                 str(llm_termination))
    return llm_termination

  def _names_from_validated_receiver(self, receiver: int, speaker: int
                                     ) -> Tuple[Tuple[str, str, str], int]:
    """Modify receiver if sending to self. Then return names of all roles.

    Args:
      receiver: integer action indicating receiver to send message to
      speaker: integer representing current message sender
    Returns:
      names: tuple of strings, (speaker_name, receiver_name, others_names)
      receiver: integer representing validated receiver
    """
    if (receiver >= self.get_game().num_players()
        or speaker >= self.get_game().num_players()):
      logging.info('Warning: rolling receiver/speaker to valid id.')
    receiver = receiver % self.get_game().num_players()
    speaker = speaker % self.get_game().num_players()
    # overwrite speaking to self as speaking to all in header
    receiver_name = ''
    if receiver == speaker:
      if len(self._names) > 2:
        receiver_name = ALL_PLAYERS
        receiver = -1
      else:
        receiver = (receiver + 1) % self.get_game().num_players()
    speaker_name = ''
    others = []
    for idx, name in enumerate(self._names):
      if idx == speaker:
        speaker_name = name
      elif idx == receiver:
        receiver_name = name
      elif receiver > -1:
        others.append(name)
    others_names = ', '.join(others)
    names = (speaker_name, receiver_name, others_names)
    return names, receiver

  def _legal_actions(self, player: int) -> List[int]:
    """Returns a list of legal actions, sorted in ascending order."""
    assert player >= 0
    return list(range(int(np.prod(self._num_actions))))

  def _apply_action(self, action: int):
    """Reply to dialogue (for agents).

    Unravel action into to tuple (who to speak to, seed to use, etc.). Then
    simulate action.

    Args:
      action: int
    """
    if self.is_chance_node():
      if self._game_setup:
        # action is an index into the list of seeds
        # use this to write the message for the previous player
        seed = self._llm_seeds[action]
        assert self._player_action is not None
        self._player_action = self._player_action or 0
        self._played_actions.append(self._player_action)
        speaker_msg = self.action_to_msg(action=self._player_action, seed=seed)
        self._apply_msg(speaker_msg)
        if self.get_game().llm_termination_prompt:
          self._llm_termination = self._llm_is_terminal()
      else:
        self._setup_game(self._init_state_configs[action])
        self._game_setup = True
    else:
      # record the action and save it to be played at chance node
      self._player_action = action
      self._current_speaker = int(self._current_player)
      self._num_actions_played += 1

  def _apply_msg(self, speaker_msg: str):
    """Update dialogue history, increment curr player, and update is_terminal.

    Args:
      speaker_msg: str
    """
    logging.info('Speaker message:\n%s', speaker_msg)
    self._dialogue.append(speaker_msg)
    self._speakers.append(self._current_player)

    # increment the current player
    self._current_player = (
        self._current_player + 1
    ) % self.get_game().num_players()

    self._player_action = None
    if self.get_game().llm_termination_prompt:
      self._llm_termination = self._llm_is_terminal()

  def apply_msg(self, speaker_msg: str):
    """Reply to dialogue (for human players and interventions).

    Args:
      speaker_msg: str
    """
    self._num_actions_played += 1
    self._played_actions.append(-1)  # assign -1 for human messages
    self._apply_msg(speaker_msg)

  def action_to_prompt(self,
                       action: int,
                       seed: int,
                       header: header_utils.Header
                       ) -> Tuple[str, str]:
    """Unravel action int to multidimensional action tuple and construct prompt.

    Args:
      action: int, the action taken in the game
      seed: int, llm seed
      header: header_utils.Header, used to format a prompt
    Returns:
      prompt: str, formatted prompt to feed the LLM to generate a new message
      header_plain: str, the formatted header without any private info / actions
    """
    speaker = int(self._current_speaker)
    action_dict = self.unravel_flat_action_to_dict(speaker, action)
    receiver = action_dict['receiver']
    opts = {**action_dict['action'], **action_dict['info']}

    names, _ = self._names_from_validated_receiver(receiver, speaker)
    speaker_name, receiver_name, others_names = names
    header_plain = header.plain.format(sender=speaker_name,
                                       receiver=receiver_name,
                                       others=others_names)

    header_w_opts = header.w_opts.format(sender=speaker_name,
                                         receiver=receiver_name,
                                         others=others_names,
                                         **opts)
    # provide header with opts to llm for response
    logging.info('Generating message (speaker=%d:%s)...',
                 speaker,
                 speaker_name)

    prompt = header.context + '\n\n' + self.dialogue_str + header_w_opts

    return prompt, header_plain

  def action_to_msg(self, action: int, seed: int) -> str:
    """Unravel action int to multidimensional action tuple and construct msg.

    Args:
      action: int, the action taken in the game
      seed: int, llm seed
    Returns:
      speaker_msg: str
    """
    header = self.get_game().header
    prompt, header_plain = self.action_to_prompt(action, seed, header)
    logging.info('LLM prompt:\n%s', prompt)

    response = self.get_game().generate_response(
        prompt=prompt,
        seed=seed,
        num_output_tokens=LLM_LENGTH_MESSAGE_TOKENS
    )
    response = response[:LLM_LENGTH_MESSAGE_CHARS]
    logging.info('LLM response:\n%s', response)

    first_special_char = text.first_special_char(
        response, len(response), self.get_game().header.special_chars)
    speaker_msg = header_plain + response[:first_special_char]

    return speaker_msg

  def unravel_flat_action_to_dict(self, speaker: int, action: int
                                  ) -> Dict[str, Any]:
    receiver, *extra_action_idxs = self._unravel_flat_action(action)

    extra_action_strs = [pa[i] for i, pa in zip(extra_action_idxs,
                                                self._prompt_actions.values())]
    action_dict = dict(zip(self.get_game().header.action_keys,
                           extra_action_strs))

    extra_info_strs = [
        pi[speaker] for pi in self._private_info.values()
    ]
    info_dict = dict(zip(self.get_game().header.info_keys, extra_info_strs))

    return {'receiver': receiver,
            'info': info_dict,
            'action': action_dict}

  def compute_rewards(self, dialogue: str) -> np.ndarray:
    """Compute rewards for each player from a given dialogue string.

    Args:
      dialogue: str, a single string with the entire dialogue thus far
    Returns:
      rewards: np.ndarray, len-num_players vector of floats
    """
    ct.set_color(logging_utils.GREEN)

    rewards = np.zeros(self.get_game().num_players(), dtype=float)

    if (not self.is_terminal() and
        self.get_game().reward_type == pyspiel.GameType.RewardModel.TERMINAL):
      return rewards

    # gather private info to compute true underlying rewards
    info_prefix = []
    for player, name in enumerate(self._names):
      extra_info_strs = [pi[player] for pi in self._private_info.values()]
      info_prefix_p = [
          f'{k}:\n{v}' for k, v in zip(self.get_game().header.info_keys,
                                       extra_info_strs)
      ]
      info_prefix_p = name + '\n' + '\n'.join(info_prefix_p)
      info_prefix.append(info_prefix_p)
    info_prefix = '\n\n'.join(info_prefix)

    # compute rewards
    for player, name in enumerate(self._names):
      player_payoffs = []
      for p, payoff in enumerate(self.get_game().payoffs):
        if payoff.obs_trans_prefix or payoff.obs_trans_postfix:
          payoff_obs_prompt = (payoff.obs_trans_prefix +
                               dialogue +
                               payoff.obs_trans_postfix)
          logging.info(ct.color('Scoring payoff (speaker=%d:%s)...'),
                       player, name)
          logging.info(ct.color('LLM prompt:\n%s'), payoff_obs_prompt)
          response = self.get_game().generate_response(
              prompt=payoff_obs_prompt,
              seed=DEFAULT_LLM_SEED,
              num_output_tokens=LLM_LENGTH_PAYOFF_OBS_TOKENS
          )
          payoff_obs = response[:LLM_LENGTH_PAYOFF_OBS_CHARS]
        else:
          payoff_obs = dialogue
        payoff_obs = info_prefix + '\n\n' + payoff_obs
        query = self._build_payoff_query(payoff.query, payoff_obs, name)
        logging.info(ct.color('Calculating payoff %d (player=%d:%s)...'),
                     p, player, name)
        logging.info(ct.color('LLM prompt:\n%s'), query)
        response = self.get_game().generate_response(
            prompt=query,
            seed=DEFAULT_LLM_SEED,
            num_output_tokens=LLM_LENGTH_SCORE_TOKENS
            )
        logging.info(ct.color('LLM response:\n%s'), response)

        logging.info(ct.color('Extracting payoff %d (player=%d:%s)...'),
                     p, player, name)
        query = (f'Extract out the final value for {name} as a single ' +
                 'numeric value from the following payoff valuation. Do ' +
                 'NOT show your work:\n\n' +
                 f'{response}\n\nResult: ')
        logging.info(ct.color('LLM prompt:\n%s'), query)
        response = self.get_game().generate_response(
            prompt=query,
            seed=DEFAULT_LLM_SEED,
            num_output_tokens=LLM_LENGTH_SCORE_TOKENS
            )
        logging.info(ct.color('LLM response:\n%s'), response)

        player_payoff = 0  # payoff defaults to 0 if LLM parsing fails
        if text.retrieve_numeric_block(response):
          player_payoff = int(text.retrieve_numeric_block(response))
          player_payoff = min(max(player_payoff, payoff.min), payoff.max)
        else:
          logging.warning(
              ct.color('Payoff extraction from response failed:\n\n%s.'),
              response)
        logging.info(ct.color('Extracted integer payoff (%s): %d'),
                     name, player_payoff)
        player_payoffs.append(player_payoff)
      rewards[player] = self.get_game().aggregate_payoffs(player_payoffs)

    ct.reset()

    return rewards.astype(float)

  def current_player(self) -> int:
    """Returns id of the next player to move, or TERMINAL if game is over."""
    if self.is_terminal():
      return pyspiel.PlayerId.TERMINAL
    elif (self._player_action is not None or  # if int, LLM msg is to be sampled
          not self._game_setup):
      return pyspiel.PlayerId.CHANCE
    else:
      return self._current_player

  def is_terminal(self) -> bool:
    """Returns True if the game is over."""
    if ((self._num_actions_played < self.get_game().max_game_length())
        and not self._llm_termination):
      return False
    else:
      return True

  def chance_outcomes(self):
    """Returns the possible chance outcomes and their probabilities."""
    assert self.is_chance_node()
    if self._game_setup:
      outcomes = range(self.get_game().num_llm_seeds)
    else:
      outcomes = range(self.get_game().num_init_states)
    p = 1.0 / len(outcomes)
    return [(o, p) for o in outcomes]

  def _action_to_string(self, player, action):
    """Action -> string."""
    if player == pyspiel.PlayerId.CHANCE:
      if self._game_setup:
        return f'Sampled LLM seed: {action}'
      else:
        return f'Sampled init state: {action}'
    else:
      action_unraveled = self.unravel_flat_action_to_dict(player, action)
      action_dict = action_unraveled['action']
      return f'Action:\nint: {action}\ndict: {action_dict}'

  def returns(self) -> np.ndarray:
    """Total reward for each player over the course of the game so far."""
    if not self.is_terminal():
      return np.zeros(self.get_game().num_players(), dtype=float)
    else:
      if self._returns is None:
        self._returns = self.compute_rewards(self.dialogue_str)
      return self._returns

  @property
  def dialogue(self) -> List[str]:
    return self._dialogue

  @property
  def dialogue_str(self) -> str:
    return ''.join(self._dialogue)

  @property
  def private_info(self) -> Dict[str, List[str]]:
    return self._private_info

  @property
  def header(self) -> header_utils.Header:
    return self.get_game().header

  @property
  def vectorize(self) -> ...:
    return self.get_game().vectorize

  @property
  def obs(self) -> List[observation_utils.Observation]:
    return self.get_game().obs

  @property
  def names(self) -> List[str]:
    """Returns list of str."""
    return self._names

  @property
  def speakers(self) -> List[int]:
    return self._speakers

  @property
  def played_actions(self) -> List[int]:
    return self._played_actions

  @property
  def num_actions(self) -> Tuple[int, ...]:
    return self._num_actions

  @property
  def prompt_actions(self) -> OrderedDict[str, List[str]]:
    return self._prompt_actions


class ChatGameObserverBase:
  """Observer, conforming to the PyObserver interface (see observation.py)."""

  def __init__(self,
               iig_obs_type: pyspiel.IIGObservationType,
               params: Union[Dict[str, Any], None]):
    """Initializes an empty observation tensor.

    Args:
      iig_obs_type: a pyspiel.IIGObservationType
      params: unused
    """
    if params:
      raise ValueError(f'Observation parameters not supported; passed {params}')

    self.iig_obs_type = iig_obs_type
    if self.iig_obs_type.perfect_recall:
      self._str_to_info_state_built = self._build_str_to_info_state()
    else:
      self._str_to_info_state_built = False

    # Determine which observation pieces we want to include.
    pieces = [('player_id', MAX_PLAYERS, (MAX_PLAYERS,))]
    if iig_obs_type.private_info == pyspiel.PrivateInfoType.SINGLE_PLAYER:
      if iig_obs_type.perfect_recall:
        pieces.append(('private_info',
                       LLM_LENGTH_MESSAGE_CHARS,
                       (LLM_LENGTH_MESSAGE_CHARS,)))
      else:
        pieces.append(('private_info', VEC_SIZE, (VEC_SIZE,)))
    if iig_obs_type.public_info:
      if iig_obs_type.perfect_recall:
        max_msgs = MAX_PLAYERS * MAX_NUM_REPLIES
        pieces.append(('scenario_prompt',
                       LLM_LENGTH_MESSAGE_CHARS,
                       (LLM_LENGTH_MESSAGE_CHARS)))
        pieces.append(('senders',
                       max_msgs * MAX_PLAYERS,
                       (max_msgs, MAX_PLAYERS)))
        pieces.append(('receivers',
                       max_msgs * MAX_PLAYERS,
                       (max_msgs, MAX_PLAYERS)))
        # record prompt actions as lossless tokenization since we do not know
        # how many actions a game will be defined with. alternatively, we could
        # record the action integer and require the user to unravel the integer
        # on the policy network side. for now, we assume the prompt action is at
        # most LLM_LENGTH_MESSAGE_CHARS subwords. we also assume everyone can
        # see everyone's actions.
        pieces.append(('prompt_actions',
                       max_msgs * LLM_LENGTH_MESSAGE_CHARS,
                       (max_msgs, LLM_LENGTH_MESSAGE_CHARS)))
        pieces.append(('messages',
                       max_msgs * LLM_LENGTH_MESSAGE_CHARS,
                       (max_msgs, LLM_LENGTH_MESSAGE_CHARS)))
      else:
        pieces.append(('dialogue', VEC_SIZE, (VEC_SIZE,)))

    # Build the single flat tensor.
    total_size = sum(size for _, size, _ in pieces)
    self.tensor = np.zeros(total_size, np.float32)

    # Build the named & reshaped views of the bits of the flat tensor.
    self.dict = {}
    index = 0
    for name, size, shape in pieces:
      self.dict[name] = self.tensor[index:index + size].reshape(shape)
      index += size

  def _build_str_to_info_state(self) -> bool:
    """Initializes map from str to infostate. Returns True if successful."""
    # Build a string tokenizer here
    # --------------------------- #
    # Build a string tokenizer here
    return True

  def _info_state(self, input_text: str, obs_size: int) -> np.ndarray:
    """Returns a len-obs_size np.ndarray given an input string and obs_size."""
    if not self._str_to_info_state_built:
      raise ValueError('String to info state mapping not built!')
    del input_text
    # Vectorize a str (ideally lossless for info state) using a tokenizer here
    # ---------------------------------------------------------------------- #
    # Vectorize a str (ideally lossless for info state) using a tokenizer here
    return np.zeros(obs_size, dtype=np.int32)

  def set_from(self, state: ChatGameState, player: int):
    """Updates `tensor` and `dict` to reflect `state` from PoV of `player`."""
    ct.set_color(logging_utils.PURPLE)

    self.tensor.fill(0)
    self.dict['player_id'][player] = 1

    extra_info_strs = [pi[player] for pi in state.private_info.values()]
    info_prefix = [
        f'{k}:\n{v}' for k, v in zip(state.header.info_keys, extra_info_strs)
    ]
    info_prefix = '\n'.join(info_prefix)
    if 'private_info' in self.dict:
      if self.iig_obs_type.perfect_recall:
        private_info = self._info_state(info_prefix, LLM_LENGTH_MESSAGE_CHARS)
      else:
        private_info = state.vectorize(info_prefix, VEC_SIZE)
      self.dict['private_info'] = private_info

    if self.iig_obs_type.public_info and self.iig_obs_type.perfect_recall:
      self.dict['scenario_prompt'] = self._info_state(state.dialogue[0],
                                                      LLM_LENGTH_MESSAGE_CHARS)
      for i, (speaker, played_action) in enumerate(zip(state.speakers,
                                                       state.played_actions)):
        self.dict['senders'][i][speaker] = 1
        if played_action >= 0:  # played_action = -1 indicates human player
          action_dict = state.unravel_flat_action_to_dict(speaker,
                                                          played_action)
          self.dict['receivers'][i][action_dict['receiver']] = 1
          pa = action_dict['action']
          action_str = '\n'.join([f'{k}: {v}' for k, v in pa.items()])
          self.dict['prompt_actions'][i] = self._info_state(
              action_str, LLM_LENGTH_MESSAGE_CHARS)

        self.dict['messages'][i] = self._info_state(state.dialogue[i + 1],
                                                    LLM_LENGTH_MESSAGE_CHARS)

    if 'dialogue' in self.dict:
      obs_prompt = (state.obs[player].obs_trans_prefix +
                    state.dialogue_str +
                    state.obs[player].obs_trans_postfix)
      logging.info(ct.color('Generating observation (speaker=%d:%s)...'),
                   player,
                   state.names[player])
      logging.info(ct.color('LLM prompt:\n%s'), obs_prompt)
      response = state.get_game().generate_response(
          prompt=obs_prompt,
          seed=DEFAULT_LLM_SEED,
          num_output_tokens=LLM_LENGTH_OBS_TOKENS
          )
      logging.info(ct.color('LLM response:\n%s'), response)
      obs = response[:LLM_LENGTH_OBS_CHARS]

      obs = info_prefix + '\n' + obs

      logging.info(ct.color('Observation (speaker=%d:%s):\n%s'),
                   player,
                   state.names[player],
                   obs)
      logging.info(ct.color('Vectorizing observation...'))
      observation = state.vectorize(obs, VEC_SIZE)
      logging.info(ct.color('Vectorized observation (speaker=%d:%s):\n%s'),
                   player,
                   state.names[player],
                   observation)
      self.dict['dialogue'] = observation

      ct.reset()

  def string_from(self, state: ChatGameState, player: int) -> str:
    """Observation of `state` from the PoV of `player`, as a string."""
    ct.set_color(logging_utils.PURPLE)

    extra_info_strs = [pi[player] for pi in state.private_info.values()]
    info_prefix = [
        f'{k}:\n{v}' for k, v in zip(state.header.info_keys, extra_info_strs)
    ]
    info_prefix = '\n'.join(info_prefix)

    if self.iig_obs_type.perfect_recall:
      return info_prefix + '\n\nFull Dialogue\n\n' + state.dialogue_str
    else:
      obs_prompt = (state.obs[player].obs_trans_prefix +
                    state.dialogue_str +
                    state.obs[player].obs_trans_postfix)
      logging.info(ct.color('Generating observation (speaker=%d:%s)...'),
                   player,
                   state.names[player])
      logging.info(ct.color('LLM prompt:\n%s'), obs_prompt)
      response = state.get_game().generate_response(
          prompt=obs_prompt,
          seed=DEFAULT_LLM_SEED,
          num_output_tokens=LLM_LENGTH_OBS_TOKENS
          )
      logging.info(ct.color('LLM response:\n%s'), response)
      obs = response[:LLM_LENGTH_OBS_CHARS]

      obs = info_prefix + '\n' + obs

      obs_str = 'Observation (speaker={:d}:{:s}):\n{:s}'.format(
          player, state.names[player], obs)

      ct.reset()

      return obs_str


class BaseChatGame(pyspiel.Game):
  """Base Chat game."""

  # pylint:disable=dangerous-default-value
  def __init__(
      self,
      params: Dict[str, Any] = DEFAULT_PARAMS,
  ):
    """Constructor.

    BaseChatGame is meant to be inherited from. Do not call its init directly.

    Args:
      params: dict, parameter dict with the following keys

        num_distinct_actions- int, # of actions at each info set
        num_llm_seeds- int, # of seeds to use for generating LLM response
        num_init_states- int, # of game setups / configurations / scenarios
        num_players- int, # of speakers (action: recipient) on the message chain
        players- int, # of speakers (action: recipient) on the message chain
          OPTIONAL. ONLY USED FOR INTERNAL OPEN_SPIEL TESTING!
        min_utility- float, minimum utility any player can attain
        max_utility- float, maximum utility any player can attain
        num_max_replies- int, total # of messages each player can send in an
          episode
    """
    if 'silence_logging' in params and params['silence_logging']:
      logging.set_verbosity(logging.ERROR)  # silence internal game logging
    self._num_distinct_actions = params['num_distinct_actions']
    if params['players'] > 0:
      logging.warning('Only meant for open_spiel testing!')
      num_players = params['players']
      self._num_players = num_players
    else:
      self._num_players = params['num_players']
    self._num_llm_seeds = params['num_llm_seeds']
    self._num_init_states = params['num_init_states']
    self._min_utility = params['min_utility']
    self._max_utility = params['max_utility']
    self._num_max_replies = params['num_max_replies']
    if params['num_max_replies'] > MAX_NUM_REPLIES:
      raise ValueError(
          f'num_max_replies ({self._num_max_replies}) exceeds ' +
          f'MAX_NUM_REPLIES ({MAX_NUM_REPLIES})')

    self._max_game_length = self._num_max_replies * self._num_players

    self._game_info = pyspiel.GameInfo(
        num_distinct_actions=self._num_distinct_actions,
        max_chance_outcomes=max(self._num_llm_seeds, self._num_init_states),
        num_players=self._num_players,
        min_utility=self._min_utility,
        max_utility=self._max_utility,
        max_game_length=self._max_game_length)

  def _load_chat_game(self,
                      observations: List[observation_utils.Observation],
                      vectorize: ...,
                      header: header_utils.Header,
                      payoffs: List[payoff_utils.Payoff],
                      aggregate_payoffs: Callable[[List[int]], float] = np.mean,
                      given_names: Union[List[str], None] = None,
                      given_llm_seeds: Union[List[int], None] = None,
                      given_prompt_actions: Union[OrderedDict[str, List[str]],
                                                  None] = None,
                      given_private_info: Union[OrderedDict[str, List[str]],
                                                None] = None,
                      initial_scenario: Union[Any, None] = None,
                      num_names: int = 2,
                      num_prompt_actions: Tuple[int, ...] = (4,),
                      num_private_info: Tuple[int, ...] = (4,),
                      examples_names: Union[List[str], None] = None,
                      examples_prompt_actions: Union[OrderedDict[str,
                                                                 List[str]],
                                                     None] = None,
                      examples_private_info: Union[OrderedDict[str, List[str]],
                                                   None] = None,
                      examples_scenarios: Union[List[Any], None] = None,
                      llm_list_suffix: str = 'Continue the list from here.',
                      llm_termination_prompt: Union[term_utils.Termination,
                                                    None] = None,
                      seed: Union[int, None] = None
                      ):
    """Constructor.

    Args:
      observations: List of Observation items used for prompting llms to extract
        observations (string features) from dialogues
      vectorize: converts any length string into a length obs_size vector

      header: List of Header items used for prompting llms to take actions
        (construct messages) based on latent action variables and private
        information

      payoffs: list of Payoff items used for constructing queries and scoring
        dialogue for each agent
      aggregate_payoffs: function that maps from vector to nonnegative scalar

      given_names: list of strings representing names of players
      given_llm_seeds: list of ints to seed llm with to generate each message
      given_prompt_actions: ordered dict mapping action_keys
        (see envs/utils/header) to list of strings representing the set of
        available prompt actions (e.g., personalities or msg tones). Overrides
        examples_prompt_actions.
      given_private_info: ordered dict mapping info_keys
        (see envs/utils/header) to length-[num_players] list of strings
        representing the private information available to each player (e.g.,
        inventory / valuations of fruits). Overrides examples_private_info.
      initial_scenario: Scenario item representing an initial message

      num_names: int, # of names to generate (can be greater than # of players)
      num_prompt_actions: tuple of int, # of prompts to consider for each
        action_key (i.e., size of action space for each prompt action)
      num_private_info: tuple of int, # of private info states to consider for
        each info_key

      examples_names: list of strings representing examples of names of players
      examples_prompt_actions: ordered dict mapping action_keys
        (see envs/utils/header) to list of strings representing examples of
        prompt actions (e.g., personalities or msg tones).
      examples_private_info: ordered dict mapping info_keys
        (see envs/utils/header) to list of strings representing examples of
        private information available to players (e.g., inventory / valuations
        of fruits). Overrides examples_private_info.
      examples_scenarios: list of Scenario items used for meta-generating new
        scenarios

      llm_list_suffix: str, gets appended to a prompt to induce an llm to
        generate a list of items (different llms like different prompts).
        chinchilla likes ``, llmit likes `Continue the list from here.`
      llm_termination_prompt: Termination item w/ [attrs query,
        obs_trans_postfix, postfix]. llm will be asked to score a binary
        response `yes`/`no` given query.format(msg=last_msg) to determine
        whether the episode has reached a terminal state (e.g., deal has been
        agreed upon). default is empty string in which case llm terminal
        condition is left unused and episode terminates after
        num_players * num_max_replies

      seed: int, master seed for experiment (used to generate all subsequent
        seeds for any random generation)
    """
    self._obs = observations
    self._vectorize = vectorize

    self._header = header

    self._payoffs = payoffs
    self._aggregate_payoffs = aggregate_payoffs
    self._max_score = aggregate_payoffs([p.max for p in payoffs])
    self._reward_type = REWARD_MODEL

    self._given_names = given_names
    self._given_llm_seeds = given_llm_seeds
    self._given_prompt_actions = given_prompt_actions
    self._given_private_info = given_private_info
    self._initial_scenario = initial_scenario

    self._num_names = max(num_names, self._num_players)
    self._num_prompt_actions = num_prompt_actions
    self._num_private_info = num_private_info

    self._examples_names = examples_names
    self._examples_prompt_actions = examples_prompt_actions
    self._examples_private_info = examples_private_info
    self._examples_scenarios = examples_scenarios

    self._llm_list_suffix = llm_list_suffix
    if llm_termination_prompt:
      query = llm_termination_prompt.query
      parsed = next(iter(string.Formatter().parse(query)), '')
      if not parsed or parsed[1] != 'msg':
        raise ValueError('Invalid llm_termination_prompt: ' +
                         f'{query}. It must include a ' +
                         'single formatting kwarg {msg}')
    self._llm_termination_prompt = llm_termination_prompt

    self._seed = seed
    self._rnd = np.random.RandomState(seed)

    if self._given_names:
      if len(self._given_names) != self._num_players:
        raise ValueError('Number of given_names does not match num_players!')
      self._names = self._given_names
      self._names_gen = False
    else:
      retrieve_name = text.retrieve_alpha_block
      self._names = self.generate_prompts('name',
                                          self._examples_names,
                                          self._num_names,
                                          retrieve_name)
      logging.info(ct.color('Generated names:\n%s', logging_utils.YELLOW),
                   '\n'.join(self._names))  # pylint:disable=logging-too-many-args
      if len(self._names) < self._num_players:
        raise ValueError(f'Generated too few names! {len(self._names)} < ' +
                         f'{self._num_players}.')
      self._names_gen = True

    if self._given_llm_seeds:
      if len(self._given_llm_seeds) != self._num_llm_seeds:
        raise ValueError('Number of given_llm_seeds does not match ' +
                         'num_llm_seeds!')
      self._llm_seeds = self._given_llm_seeds
      self._llm_seeds_gen = False
    else:
      self._llm_seeds = list(self._rnd.randint(MIN_RND_SEED, MAX_RND_SEED,
                                               size=self._num_llm_seeds))
      logging.info(ct.color('Generated action seeds:%s', logging_utils.YELLOW),
                   self._llm_seeds)  # pylint:disable=logging-too-many-args
      self._llm_seeds_gen = True

    # loop over every action key in header action keys
    # if action key is in given prompt action, use it and overwrite
    # else, generate it
    def retrieve_prompt(llm_response: str) -> str:
      useless_chars = (' ', '\n')
      special_chars = ITEM_PREFIX
      for char in useless_chars:
        special_chars = special_chars.strip(char)
      special_chars = tuple(special_chars)
      return text.retrieve_special_char_block(llm_response,
                                              special_chars=special_chars,
                                              useless_chars=useless_chars)

    prompt_action_lists = []
    if not self._header.action_keys:
      self._num_prompt_actions = tuple([])
    for i, action_key in enumerate(self._header.action_keys):
      if (self._given_prompt_actions and
          action_key in self._given_prompt_actions):
        action_list = self._given_prompt_actions[action_key]
        if len(action_list) != self._num_prompt_actions[i]:
          logging.info(ct.color(f'Overwriting num_prompt_actions[{i}]=' +
                                f'{self._num_prompt_actions[i]} to reflect ' +
                                f'given len-{len(action_list)} prompt ' +
                                f'action list for action_key={action_key}.',
                                color=logging_utils.YELLOW))
          if isinstance(self._num_prompt_actions, tuple):
            self._num_prompt_actions = list(self._num_prompt_actions)
          self._num_prompt_actions[i] = len(action_list)
      else:
        examples = self._examples_prompt_actions[action_key]
        action_list = self.generate_prompts(action_key,
                                            examples,
                                            self._num_prompt_actions[i],
                                            retrieve_prompt)
        logging.info(ct.color(
            'Generated prompt actions for action key = %s:\n%s',
            color=logging_utils.YELLOW),
                     action_key, '\n-----\n'.join(action_list))
      prompt_action_lists.append(action_list)
    self._prompt_actions = collections.OrderedDict(zip(self._header.action_keys,
                                                       prompt_action_lists))
    if isinstance(self._num_prompt_actions, list):
      self._num_prompt_actions = tuple(self._num_prompt_actions)

    if (self._initial_scenario
        and self._given_private_info
        and tuple(self._given_private_info.keys()) != self._header.info_keys):
      raise ValueError('Must define private info for each player if setting' +
                       ' an initial scenario.')

    private_info_lists = []
    if not self._header.info_keys:
      self._num_private_info = tuple([])
    for i, info_key in enumerate(self._header.info_keys):
      if self._given_private_info and info_key in self._given_private_info:
        info_list = self._given_private_info[info_key]
        if self._initial_scenario:
          if len(info_list) < self._num_players:
            raise ValueError('Must define at least a single private info for ' +
                             'each player if setting an initial scenario. ' +
                             f'Num_players={self._num_players} but only given' +
                             f' len-{len(info_list)} private info list for ' +
                             f'info_key={info_key}.')
          else:
            info_list = info_list[:self._num_players]
        if len(info_list) != self._num_private_info[i]:
          logging.info(ct.color(f'Overwriting num_private_info[{i}]=' +
                                f'{self._num_private_info[i]} to reflect ' +
                                f'given len-{len(info_list)} private info ' +
                                f'list for info_key={info_key}.',
                                color=logging_utils.YELLOW))
          if isinstance(self._num_private_info, tuple):
            self._num_private_info = list(self._num_private_info)
          self._num_private_info[i] = len(info_list)
      else:
        examples = self._examples_private_info[info_key]
        info_list = self.generate_prompts(info_key,
                                          examples,
                                          self._num_private_info[i],
                                          retrieve_prompt)
        logging.info(ct.color('Generated private info for info key = %s:\n%s',
                              color=logging_utils.YELLOW),
                     info_key, '\n-----\n'.join(info_list))
      private_info_lists.append(info_list)
    self._private_info = collections.OrderedDict(zip(self._header.info_keys,
                                                     private_info_lists))
    if isinstance(self._num_private_info, list):
      self._num_private_info = tuple(self._num_private_info)

    if self._examples_scenarios:
      self._meta_query = self._build_meta_query(self._examples_scenarios)
    else:
      self._meta_query = None

    if self._initial_scenario:
      valid = self._initial_scenario_is_valid(self._initial_scenario)
      assert valid, ('Scenario does not match given game spec (names, actions' +
                     ', info, ...')
      self._initial_scenario = self._initial_scenario
    else:
      self._initial_scenario = None

    self._num_actions = (
        self._num_players,
    ) + tuple(self._num_prompt_actions)

    na = int(np.prod(self._num_actions))
    if na != self._num_distinct_actions:
      raise ValueError(f'Size of prompt action space ({na}) does not match ' +
                       f'num_distinct_actions ({self._num_distinct_actions})!')

    self._initial_state_configs = self.new_initial_state_configs()

  def _generate_response(self, prompt: str, seed: int,
                         num_output_tokens: Union[int, None] = None) -> str:
    """Returns LLM generated string given prompt and seed."""
    return ''

  def _generate_bool(self, prompt: str, seed: int) -> bool:
    """Returns LLM generated boolean given prompt and seed."""
    return False

  def _build_meta_query(self, scenarios=List[Tuple]) -> str:
    """Build prompt with several scenarios for generating new scenarios."""
    wrapped_scenarios = []
    for s in scenarios:
      scenario_header_unformatted = self._header.w_opts + s.msg
      s_asdict = dataclasses.asdict(s)
      scenario_header = scenario_header_unformatted.format(**s_asdict,
                                                           others=ALL_PLAYERS)
      wrapped_scenarios.append(scenario_header)
    return ''.join(wrapped_scenarios)

  def _initial_scenario_is_valid(self, scenario: Any) -> bool:
    """Check all components of scenario are well defined and return bool."""
    fields = list(scenario.__dataclass_fields__.keys())

    req_fields = ['sender', 'receiver'] + list(self._header.action_keys)
    req_fields += list(self._header.info_keys)
    valid_fields = True
    for req_field in req_fields:
      valid_fields = (valid_fields and req_field in fields)

    if not valid_fields:
      raise ValueError(f'Scenario must define required fields: {req_fields}. ' +
                       f'Found fields: {fields}')

    valid_players = (scenario.sender in self._names
                     and scenario.receiver in self._names + [ALL_PLAYERS])

    scenario_dict = dataclasses.asdict(scenario)

    valid_actions = True
    for key in self._header.action_keys:
      valid_actions = (valid_actions and
                       key in scenario_dict and
                       scenario_dict[key] in self._prompt_actions[key])

    valid_info = True
    for key in self._header.info_keys:
      # private_info[key][i] is unique to player i
      # initial scenario contains player 0's private info and must match the
      # first item in the list of private information provided
      valid_info = (valid_info and
                    key in scenario_dict and
                    scenario_dict[key] == self._private_info[key][0])

    valid = valid_players and valid_actions and valid_info

    return valid

  def generate_prompts(self, key, examples, num_prompts,
                       retrieve_prompt: Callable[[str], str]) -> List[str]:
    """Generates a list of distinct prompts from an initial list.

    Args:
      key: str, (descriptive) name of prompt type
      examples: list of str, example prompts to seed llm
      num_prompts: int, number of distinct prompts to generate
      retrieve_prompt: function to retrieve example from string

    Returns:
      prompts: list of strings
    """
    ct.set_color(logging_utils.CYAN)

    answers = set()
    num_gen = LLM_LIST_GEN_ATTEMPTS
    prompt = ['#### INSTRUCTIONS #####',
              'Given a list of items from a given category, continue the list' +
              ' and generate an additional item from the same category. The ' +
              f'category is {key}s. Use `{ITEM_PREFIX}` to denote separate ' +
              'items. Do not include the list of items from the input prompt ' +
              'in your output response.']
    prompt = '\n'.join(text.wrap(prompt)) + '\n'
    prompt += ('Input:\n' + ITEM_PREFIX +
               ('\n' + ITEM_PREFIX).join(examples) + '\n' +
               self._llm_list_suffix)
    logging.info(ct.color('Generating list of distinct prompts...'))
    logging.info(ct.color('Example prompt:\n%s'), prompt)
    for seed in self._rnd.randint(MIN_RND_SEED, MAX_RND_SEED, size=num_gen):
      logging.info(ct.color('Generating %s (seed=%s)'), key, seed)
      response = self.generate_response(
          prompt=prompt,
          seed=seed,
          num_output_tokens=LLM_LENGTH_LIST_OF_WORDS_TOKENS
          )
      logging.info(ct.color('LLM response\n%s'), response)
      answer = retrieve_prompt(response)
      if answer and answer not in answers:
        answers.add(answer)
      if len(answers) >= num_prompts:
        return list(answers)
    num_distinct = len(answers)
    if len(answers) < num_prompts:
      logging.warning(ct.color(
          'Only %d distinct prompts generated for %d desired:\n%s.'),
                      num_distinct, num_prompts, answers)

    ct.reset()

    return list(answers)

  def generate_scenario(self) -> Tuple[List[str],
                                       OrderedDict[str, List[str]],
                                       Any]:
    """Generates a new game config from examples.

    Returns:
      given_names: list of str
      given_private_info: OrderedDict(str: list of str)
      initial_scenario(msg, sender, receiver, **private_info, **prompt_actions)
    """
    player_names = self._rnd.choice(self._names,
                                    size=self._num_players,
                                    replace=False)
    sender, receiver = player_names[:2]
    if self._num_players > 2:
      others = ', '.join(player_names[2:])
    else:
      others = ''

    pa_lists = self._prompt_actions.values()
    prompt_action_vals = [self._rnd.choice(pa_list) for pa_list in pa_lists]
    prompt_actions_header = collections.OrderedDict(zip(
        self._header.action_keys, prompt_action_vals))

    pi_lists = self._private_info.values()
    private_info_vals = [
        self._rnd.choice(pi_list, size=self._num_players)
        for pi_list in pi_lists
    ]
    private_info = collections.OrderedDict(zip(self._header.info_keys,
                                               private_info_vals))
    private_info_vals_player_0 = [piv[0] for piv in private_info_vals]
    private_info_header = collections.OrderedDict(zip(
        self._header.info_keys, private_info_vals_player_0))

    opts = prompt_actions_header
    opts.update(private_info_header)

    # scenarios are generated drawing from a fixed set of personalities
    header = self._header.w_opts.format(sender=sender,
                                        receiver=receiver,
                                        others=others,
                                        **opts)

    # generate a random scenario
    # need to generate new scenario with specific players (i.e. names). Can
    # 1) try to generate multiple scenarios at once and parse output
    # 2) generate a single scenario by varying the LLM seed
    # 3) can rely on the randomness in names and private info to induce new
    #    scenarios
    # we are currently going with a mix of options 2) and 3)
    logging.info('Generating initial scenario...')
    logging.info('Scenario prompt:\n%s', self._meta_query + header)
    response = self.generate_response(
        prompt=self._meta_query + header,
        seed=self._seed,
        num_output_tokens=LLM_LENGTH_MESSAGE_TOKENS
        )
    response = response[:LLM_LENGTH_MESSAGE_CHARS]
    logging.info('LLM response:\n%s', response)
    examples = []
    ptr = 0
    i = 0
    augmented_response = header + response
    while ptr < len(augmented_response):
      generated_example = self._header.strip_msg(augmented_response[ptr:],
                                                 sender)
      if not generated_example:
        break
      ptr += len(generated_example)
      generated_example = generated_example.strip('\n')
      logging.info('*Generated Example %d:\n%s', i, generated_example)
      i += 1
      examples.append(generated_example)
    # grab first generated scenario
    scenario_prompt = examples[0]
    logging.info('Example 0 selected')
    actions = collections.OrderedDict(zip(['player_names'],
                                          [player_names]))
    actions.update(self._prompt_actions)

    given_names = player_names
    given_private_info = private_info
    scenario_class = self._examples_scenarios[0].__class__
    initial_scenario = scenario_class(msg=scenario_prompt,
                                      sender=sender,
                                      receiver=receiver,
                                      **opts)

    return (given_names, given_private_info, initial_scenario)

  def new_initial_state_configs(self) -> Tuple[InitialStateConfiguration, ...]:
    """Generates a tuple of new dialogue game(s).

    Returns:
      Tuple of InitialStateConfiguration(s)
    """
    raw_setups = []
    if self._initial_scenario:
      raw_setups.append(
          (self._names, self._private_info, self._initial_scenario))
    else:
      for _ in range(self._num_init_states):
        raw_setups.append(self.generate_scenario())

    initial_state_configs = []
    for raw_setup in raw_setups:
      names, private_info, scenario = raw_setup
      scenario_prompt_unformatted = self._header.plain + scenario.msg
      scenario_prompt = scenario_prompt_unformatted.format(
          sender=scenario.sender,
          receiver=scenario.receiver,
          others=ALL_PLAYERS)
      actions = collections.OrderedDict(zip(['player_names'], [names]))
      actions.update(self._prompt_actions)
      initial_state_config = InitialStateConfiguration(
          actions=actions,
          seeds=self._llm_seeds,
          scenario_prompt=scenario_prompt,
          private_info=private_info
          )
      initial_state_configs.append(initial_state_config)

    return tuple(initial_state_configs)

  @property
  def game_info(self) -> pyspiel.GameInfo:
    return self._game_info

  @property
  def obs(self) -> List[observation_utils.Observation]:
    return self._obs

  @property
  def vectorize(self) -> Any:
    return self._vectorize

  @property
  def header(self) -> header_utils.Header:
    return self._header

  @property
  def payoffs(self) -> List[payoff_utils.Payoff]:
    return self._payoffs

  @property
  def aggregate_payoffs(self) -> Callable[[List[int]], float]:
    return self._aggregate_payoffs

  @property
  def reward_type(self) -> pyspiel.GameType.RewardModel:
    return self._reward_type

  @property
  def rnd(self) -> np.random.RandomState:
    return self._rnd

  @property
  def llm_termination_prompt(self) -> Union[term_utils.Termination, None]:
    return self._llm_termination_prompt

  @property
  def llm_seeds(self) -> List[int]:
    return self._llm_seeds

  @property
  def num_llm_seeds(self) -> int:
    return self._num_llm_seeds

  @property
  def num_init_states(self) -> int:
    return self._num_init_states

  @property
  def given_prompt_actions(self) -> Union[OrderedDict[str, List[str]], None]:
    return self._given_prompt_actions



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for parsing and constructing message strings.
"""

import textwrap

from typing import List, Tuple


def strip_msg(text: str,
              block_msg: str,
              block_opt: str,
              terminal_str: str = '') -> str:
  """Strip email message (with header) from text block, i.e., [ (A) - (B) ).

  Assumes messages adhere to the following format:
  BLOCK_OPT
  <-- action & info -->
  BLOCK_MSG (A)
  <-- e.g., sender/receiver -->
  BLOCK_MSG
  <-- e.g., message -->
  BLOCK_OPT (B)

  Args:
    text: str
    block_msg: str, string of characters delineating the message
    block_opt: str, string of characters demarking the start of
      the options (actions and info)
    terminal_str: str (optional), indicates the end of a message if block_opt
      is not found. this will be included in the stripped output.
  Returns:
    stripped_text: str
  """
  ctr = 0
  right_ptr = 0
  left_ptr = text.find(block_msg)
  if left_ptr == -1:
    return ''
  while ctr < 2:
    block_idx = text[right_ptr:].find(block_msg)
    if block_idx == -1:
      return ''
    right_ptr += block_idx + len(block_msg)
    ctr += 1
  block_idx = text[right_ptr:].find(block_opt)
  if block_idx != -1:  # if find block_opt return message ending at (B)
    right_ptr += block_idx
  else:
    if terminal_str:  # if no block_opt, return message ending at terminal_str
      block_idx = text[right_ptr:].find(terminal_str)
      if block_idx != -1:
        right_ptr += block_idx + len(terminal_str)
      else:  # if no terminal_str, return message to end of text string
        right_ptr = len(text)
  return text[left_ptr:right_ptr]


def first_special_char(text: str,
                       max_idx: int,
                       special_chars: Tuple[str, ...]) -> int:
  first_special_chars = [max_idx]
  for char in special_chars:
    idx = text.find(char)
    if idx < 0:
      first_special_chars.append(max_idx)
    else:
      first_special_chars.append(idx)
  return min(first_special_chars)


def retrieve_special_char_block(text: str,
                                special_chars: Tuple[str, ...] = ('*',),
                                useless_chars: Tuple[str, ...] = (' ', '\n')):
  for char in special_chars:
    text = text.strip(char)
  idx_end = first_special_char(text, len(text), special_chars)
  text = text[:idx_end]
  for char in useless_chars:
    text = text.strip(char)
  return text


def retrieve_alpha_block(text: str) -> str:
  """Return the first instance of a contiguous alpha(not numeric) substring."""
  first_alpha_char = next(filter(str.isalpha, text), -1)
  if first_alpha_char == -1:
    return ''
  start = text.find(first_alpha_char)
  sliced = text[start:]
  last_alpha_char = next(filter(lambda s: not str.isalpha(s), sliced), -1)
  if last_alpha_char == -1:
    return sliced
  finish = sliced.find(last_alpha_char)
  return text[start:start + finish]


def retrieve_numeric_block(text: str) -> str:
  """Return the first instance of a contiguous numeric(not alpha) substring."""
  first_numeric_char = next(filter(str.isnumeric, text), -1)
  if first_numeric_char == -1:
    return ''
  start = text.find(first_numeric_char)
  sliced = text[start:]
  last_numeric_char = next(filter(lambda s: not str.isnumeric(s), sliced), -1)
  if start > 0 and text[start - 1] == '-':
    start -= 1
    sliced = text[start:]
  if last_numeric_char == -1:
    return sliced
  finish = sliced.find(last_numeric_char)
  return text[start:start + finish]


def wrap(message: List[str]) -> List[str]:
  """Given a list of strings, returns a list of them `wrapped` (paragraphs).

  Args:
    message: list of strings
  Returns:
    wrapped: list of strings with each string `wrapped` so that each line only
      contains (default) 70 characters
  """
  wrapped = []
  for sub_msg in message:
    sub_msg_wrapped = textwrap.wrap(sub_msg)
    if len(sub_msg_wrapped) > 1:
      sub_msg_wrapped = ['\n'.join(sub_msg_wrapped)]
    wrapped.extend(sub_msg_wrapped)
  return wrapped



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Meta learning algorithm."""

import os
from typing import Dict, List, Any

from absl import flags
from absl import logging
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import optax

from open_spiel.python.examples.meta_cfr.sequential_games import cfr
from open_spiel.python.examples.meta_cfr.sequential_games import dataset_generator
from open_spiel.python.examples.meta_cfr.sequential_games import game_tree_utils
from open_spiel.python.examples.meta_cfr.sequential_games import models
from open_spiel.python.examples.meta_cfr.sequential_games import openspiel_api
from open_spiel.python.examples.meta_cfr.sequential_games import typing
from open_spiel.python.examples.meta_cfr.sequential_games import utils


FLAGS = flags.FLAGS

flags.DEFINE_integer("batch_size", 250, "Batch size.")
flags.DEFINE_integer("num_batches", 1, "Number of batches.")
flags.DEFINE_integer("meta_learner_training_epochs", 1,
                     "Number of meta_learner_training_epochs")
flags.DEFINE_integer("num_tasks", 1, "Number tasks to train meta learner.")
flags.DEFINE_integer("random_seed", 2, "Random seed.")
flags.DEFINE_integer("checkpoint_interval", 50,
                     "Checkpoint every checkpoint_interval.")
flags.DEFINE_string("game", "leduc_poker", "Name of the game")
flags.DEFINE_integer("players", 2, "Number of players")
flags.DEFINE_bool("perturbation", True, "Random perturbation of the game.")
flags.DEFINE_bool(
    "use_infostate_representation", True,
    "Use infostate representation as extra input to meta network.")
flags.DEFINE_float("init_lr", 0.2, "Initial learning rate")
flags.DEFINE_string("lstm_sizes", "64", "Size of lstm layers.")
flags.DEFINE_string("mlp_sizes", "20, 20", "Size of mlp layers.")
flags.DEFINE_string("model_type", "MLP", "Model type.")


os.environ["XLA_PYTHON_CLIENT_PREALLOCATE"] = "false"
os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = "1.5"


def append_counterfactual_values(
    infostates: List[typing.InfostateNode],
    counterfactual_values: Dict[str, List[List[float]]]):
  for infostate in infostates:
    counterfactual_values[infostate.infostate_string].append([
        infostate.counterfactual_action_values[a]
        for a in infostate.get_actions()
    ])


def compute_next_policy_invariants(
    infostates: typing.InfostateMapping, all_actions: List[int],
    infostate_map: typing.InfostateMapping
) -> tuple[Dict[str, jnp.ndarray], Dict[str, List[int]]]:
  """Computes information needed to calculate next policy.

  This function computes one hot encodings of infostates and returns mappings
  from infostate strings to one hot representations of infostates as well as
  illegal actions.

  Args:
    infostates: List of infostate mappings.
    all_actions: List of actions.
    infostate_map: Mapping from infostate string to infostate.

  Returns:
    Returns mappings of infostate strings to one hot representation for
    infostates and illegal actions
  """
  one_hot_representations = {}
  illegal_actions = {}

  for (infostate_str, infostate) in infostates.items():
    if infostate.is_terminal():
      continue

    legal_actions = infostate.get_actions()

    if len(legal_actions) == 1:
      infostate.policy[infostate.get_actions()[0]] = 1
      continue
    infostate_str_one_hot = jax.nn.one_hot(infostate_map[infostate_str],
                                           len(infostates))
    one_hot_representations[infostate_str] = infostate_str_one_hot
    illegal_actions[infostate_str] = [
        i for i, a in enumerate(all_actions) if a not in legal_actions
    ]
  return one_hot_representations, illegal_actions


def compute_next_policy(infostates: typing.InfostateMapping,
                        net_apply: typing.ApplyFn, net_params: typing.Params,
                        epoch: int, all_actions: List[int],
                        one_hot_representations: Dict[str, jnp.ndarray],
                        illegal_actions: Dict[str,
                                              List[int]], key: hk.PRNGSequence):
  """Computes next step policy from output of the model.

  Args:
    infostates: List of infostate mappings.
    net_apply: Apply function.
    net_params: Model params.
    epoch: epoch.
    all_actions: List of actions.
    one_hot_representations: Dictionary from infostate string to infostate.
    illegal_actions: Dictionary from infostate string to the list of illegal
      actions.
    key: Haiku Pseudo random number generator.
  """

  infostate_lst = []
  input_lst = []
  illegal_action_lst = []

  batched_net_output = []
  for (infostate_str, infostate) in infostates.items():
    if infostate.is_terminal():
      continue

    legal_actions = infostate.get_actions()
    if len(legal_actions) == 1:
      infostate.policy[infostate.get_actions()[0]] = 1
      continue
    regret_vec = np.array([
        infostate.regret[a] /
        (epoch + 1) if a in infostate.get_actions() else 0
        for a in all_actions
    ])
    if FLAGS.use_infostate_representation:
      one_hot_representation = one_hot_representations[infostate_str]
      net_input = jnp.concatenate([regret_vec, one_hot_representation])
    else:
      net_input = regret_vec
    input_lst.append(net_input)
    infostate_lst.append(infostate)
    illegal_action_lst.append(illegal_actions[infostate_str])
  batched_inputs, output_mappings, relevant_illegal_actions = (
      utils.get_batched_input(
          input_lst, infostate_lst, illegal_action_lst, FLAGS.batch_size
      )
  )
  idx = 0

  for _ in range(int(len(batched_inputs) / FLAGS.batch_size)):
    batched_input, output_mapping, relevant_illegal_action = batched_inputs[
        idx:idx + FLAGS.batch_size], output_mappings[
            idx:idx +
            FLAGS.batch_size], relevant_illegal_actions[idx:idx +
                                                        FLAGS.batch_size]
    idx += FLAGS.batch_size

    batched_input_jnp = jnp.array(
        np.expand_dims(np.array(batched_input), axis=1))
    batched_net_output = utils.get_network_output_batched(  # pytype: disable=wrong-arg-types  # jnp-type
        net_apply, net_params,
        batched_input_jnp,
        relevant_illegal_action, key)
    for i, infostate in enumerate(output_mapping):
      net_output = jnp.squeeze(batched_net_output[i])
      for ai, action in enumerate(infostate.get_actions()):
        infostate.policy[action] = float(net_output[ai])


def cfr_br_meta_data(
    history_tree_node: typing.HistoryNode,
    infostate_nodes: List[typing.InfostateNode],
    all_infostates_map: List[typing.InfostateMapping], epochs: int,
    net_apply: typing.ApplyFn, net_params: typing.Params,
    all_actions: List[int], infostate_map: typing.InfostateMapping,
    key: hk.PRNGSequence
) -> tuple[Dict[str, jnp.ndarray], Dict[str, jnp.ndarray], List[float]]:
  """Collects counterfactual values for both players and best response for player_2.

  Args:
    history_tree_node: Game tree HistoryTreeNode which is the root of the game
      tree.
    infostate_nodes: Infostates.
    all_infostates_map: List of mappings from infostate strings to infostates.
    epochs: Number of epochs.
    net_apply: Apply function.
    net_params: Network parameters.
    all_actions: List of all actions.
    infostate_map: A mapping from infostate strings to infostates.
    key: Haiku pseudo random number generator.

  Returns:
    Returns counterfactual values for player_1, counterfactual values for
    player_2 and best response values for player_2.
  """
  counterfactual_values_player1 = {
      infostate.infostate_string: []
      for infostate in list(all_infostates_map[1].values())
  }
  counterfactual_values_player2 = {
      infostate.infostate_string: []
      for infostate in list(all_infostates_map[2].values())
  }

  non_terminal_infostates_map_player1 = utils.filter_terminal_infostates(
      all_infostates_map[1]
  )
  one_hot_representations_player1, illegal_actions_player1 = (
      compute_next_policy_invariants(
          non_terminal_infostates_map_player1, all_actions, infostate_map
      )
  )
  player_2_last_best_response_values = []
  for epoch in range(epochs):
    compute_next_policy(non_terminal_infostates_map_player1, net_apply,
                        net_params, epoch, all_actions,
                        one_hot_representations_player1,
                        illegal_actions_player1, key)

    cfr.compute_reach_probabilities(history_tree_node, all_infostates_map)
    cfr.cumulate_average_policy(list(all_infostates_map[1].values()))
    cfr.compute_best_response_policy(infostate_nodes[2])
    cfr.compute_reach_probabilities(history_tree_node, all_infostates_map)
    cfr.compute_counterfactual_values(infostate_nodes[1])
    cfr.update_regrets(list(all_infostates_map[1].values()))
    append_counterfactual_values(
        list(all_infostates_map[1].values()), counterfactual_values_player1)
    cfr.normalize_average_policy(all_infostates_map[1].values())
    cfr.compute_reach_probabilities(history_tree_node, all_infostates_map)
    player_2_last_best_response_values.append(
        float(cfr.compute_best_response_values(infostate_nodes[2]))
    )

    logging.info(
        "Epoch %d: player_2 best response value is %f",
        epoch,
        player_2_last_best_response_values[-1],
    )

  return (  # pytype: disable=bad-return-type  # jax-ndarray
      counterfactual_values_player1,
      counterfactual_values_player2,
      player_2_last_best_response_values,
  )


class MetaCFRRegretAgent:
  """Meta regret minimizer agent.

  Attributes:
    training_epochs: Number of training epochs.
    meta_learner_training_epochs: Number of epochs for meta learner.
    game_name: Name of the game.
    game_config: Game configuration.
    perturbation: Binary variable to specify perturbation.
    seed: Random seed.
    model_type: Type of NN model for meta learner.
    best_response: Binary variable to specify if using best response.
    optimizer: Optimizer model.
  """

  def __init__(self,
               training_epochs,
               meta_learner_training_epochs,
               game_name,
               game_config,
               perturbation,
               seed,
               model_type="MLP",
               best_response=True):
    self._training_epochs = training_epochs
    self._meta_learner_training_epochs = meta_learner_training_epochs
    self._game_name = game_name
    self._model_type = model_type
    self._perturbation = perturbation
    self._game_config = game_config
    self._best_response = best_response
    self._seed = seed
    self._rng = hk.PRNGSequence(100)
    self._world_state = openspiel_api.WorldState(self._game_name,
                                                 self._game_config,
                                                 self._perturbation,
                                                 self._seed)
    self._all_actions = self._world_state.get_distinct_actions()
    self._num_infostates, self._infostate_map = self.get_num_infostates()
    self._step = 0

  def get_num_infostates(self):
    """Returns number of infostates and infostate mapping.

    Returns:
      Returns sum of number of infostates for both players and a mapping from
      infostate string to infostates.
    """
    all_infostates_map = [{}, {}, {}]
    _, _ = game_tree_utils.build_tree_dfs(
        self._world_state, all_infostates_map)
    non_terminal_infostates_map_player1 = utils.filter_terminal_infostates(
        all_infostates_map[1])
    non_terminal_infostates_map_player2 = utils.filter_terminal_infostates(
        all_infostates_map[2])
    if self._best_response:
      infostate_map = {
          infostate_str: infostate_node
          for (infostate_node, infostate_str
              ) in enumerate(list(non_terminal_infostates_map_player1.keys()))
      }
      return len(non_terminal_infostates_map_player1), infostate_map
    nont_terminal_infostates_map_both_players = list(
        non_terminal_infostates_map_player1.keys()) + list(
            non_terminal_infostates_map_player2.keys())
    infostate_map = {
        infostate_str: infostate_node
        for (infostate_node, infostate_str
            ) in enumerate(nont_terminal_infostates_map_both_players)
    }
    return len(non_terminal_infostates_map_player1) + len(
        non_terminal_infostates_map_player2), infostate_map

  def train(self):
    self.training_optimizer()

  def next_policy(self, world_state: openspiel_api.WorldState):
    """Computes best reponses for the next step of cfr.

    Args:
      world_state: Current state of the world.

    Returns:
      Returns best response values for player_2.

    """
    all_infostates_map = [{}, {}, {}]
    first_history_node, infostate_nodes = game_tree_utils.build_tree_dfs(
        world_state, all_infostates_map)

    _, _, player_2_best_response_values = cfr_br_meta_data(  # pytype: disable=wrong-arg-types
        history_tree_node=first_history_node,
        infostate_nodes=infostate_nodes,
        all_infostates_map=all_infostates_map,
        epochs=self._meta_learner_training_epochs,
        net_apply=self.optimizer.net_apply,  # pytype: disable=attribute-error
        net_params=self.optimizer.net_params,  # pytype: disable=attribute-error
        all_actions=self._all_actions,
        infostate_map=self._infostate_map,
        key=self._rng)
    return player_2_best_response_values

  def optimize_infoset(self, cfvalues: Any, infoset: List[typing.InfostateNode],
                       infostate_map: typing.InfostateMapping,
                       rng: hk.PRNGSequence):
    """Apply updates to optimizer state.

    Args:
      cfvalues: Counterfactual values.
      infoset: Infostates.
      infostate_map: Mapping from infostate string to infostate.
      rng: Next random seed.
    """
    grads = jax.grad(
        utils.meta_loss, has_aux=False)(self.optimizer.net_params, cfvalues,  # pytype: disable=attribute-error
                                        self.optimizer.net_apply,  # pytype: disable=attribute-error
                                        self._meta_learner_training_epochs,
                                        len(self._all_actions), infoset,
                                        infostate_map, FLAGS.batch_size,
                                        next(rng),
                                        FLAGS.use_infostate_representation)
    updates, self.optimizer.opt_state = self.optimizer.opt_update(  # pytype: disable=attribute-error
        grads, self.optimizer.opt_state)  # pytype: disable=attribute-error

    self.optimizer.net_params = optax.apply_updates(self.optimizer.net_params,  # pytype: disable=attribute-error
                                                    updates)

  def training_optimizer(self):
    """Train an optimizer for meta learner."""

    self.optimizer = models.OptimizerModel(
        mlp_sizes=FLAGS.mlp_sizes,
        lstm_sizes=FLAGS.lstm_sizes,
        initial_learning_rate=FLAGS.init_lr,
        batch_size=FLAGS.batch_size,
        num_actions=len(self._all_actions),
        num_infostates=self._num_infostates,
        model_type=self._model_type,
        use_infostate_representation=FLAGS.use_infostate_representation)
    self.optimizer.initialize_optimizer_model()

    while self._step < FLAGS.num_tasks:
      if self._perturbation:
        self._seed = np.random.choice(np.array(list(range(100))))
      self._world_state = openspiel_api.WorldState(
          self._game_name,
          self._game_config,
          perturbation=self._perturbation,
          random_seed=self._seed)

      for epoch in range(self._training_epochs):
        logging.info("Training epoch %d", epoch)
        all_infostates_map = [{}, {}, {}]
        first_history_node, infostate_nodes = game_tree_utils.build_tree_dfs(
            self._world_state, all_infostates_map)
        cfr_values_player1, cfr_values_player2, _ = cfr_br_meta_data(  # pytype: disable=wrong-arg-types
            history_tree_node=first_history_node,
            infostate_nodes=infostate_nodes,
            all_infostates_map=all_infostates_map,
            epochs=self._meta_learner_training_epochs,
            net_apply=self.optimizer.net_apply,
            net_params=self.optimizer.net_params,
            all_actions=self._all_actions,
            infostate_map=self._infostate_map,
            key=self._rng)

        train_dataset = []
        cfvalues_per_player = [
            cfr_values_player1, cfr_values_player2
        ]
        # for CFRBR we consider player 0.
        player_ix = 0
        infosets = [
            infoset for infoset in all_infostates_map[player_ix + 1].values()
            if len(infoset.get_actions()) >= 2
        ]
        for infoset in infosets:
          cfvalues = cfvalues_per_player[player_ix][infoset.infostate_string]
          train_dataset.append((cfvalues, infoset))

        dataset = dataset_generator.Dataset(train_dataset, FLAGS.batch_size)  # pytype: disable=wrong-arg-types  # jax-ndarray
        data_loader = dataset.get_batch()
        for _ in range(FLAGS.num_batches):
          batch = next(data_loader)
          cfvalues, infoset = zip(*batch)
          cfvalues = np.array(list(cfvalues), dtype=object)
          cfvalues = utils.mask(cfvalues, infoset, len(self._all_actions),
                                FLAGS.batch_size)
          self.optimize_infoset(cfvalues, infoset, self._infostate_map,  # pytype: disable=wrong-arg-types
                                self._rng)
      logging.info("Game: %d", self._step)
      self._step += 1



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A few prompts for extracting the value of a schedule negotiation with llms.
"""

from open_spiel.python.games.chat_games.envs.payoffs import utils as payoff_utils
from open_spiel.python.games.chat_games.envs.utils import text

MIN_PAYOFF = 0
MAX_PAYOFF = 20
PAYOFF_PROMPT = '''
An intelligent assistant is looking at summaries of dialogues between two people
trying to decide when to meet. It also observes the day preferences of
participants as well as the days the participant is out of office. It is trying
to decide how happy each participant is with the outcome of the dialogue and how
happy they are with the chosen meeting time.

Example 1:
Alice:
ooo_days:
monday: false
tuesday: true
wednesday: true
thursday: false
friday: false
saturday: true
sunday: false
day_prefs
monday: 2
tuesday: 4
wednesday: 12
thursday: 8
friday: 5
saturday: 0
sunday: 0

Bob:
ooo_days:
monday: false
tuesday: true
wednesday: true
thursday: false
friday: false
saturday: true
sunday: false
day_prefs
monday: 10
tuesday: 5
wednesday: 15
thursday: 3
friday: 2
saturday: 1
sunday: 1

Outcome Summary: Meeting agreed on Monday.

Final valuation for Bob: 10.
Calculation: Monday selected. Not an out of office day. Value of monday: 10.

Example 2:
Alice:
ooo_days:
monday: false
tuesday: true
wednesday: true
thursday: false
friday: false
saturday: true
sunday: false
day_prefs:
monday: 10
tuesday: 5
wednesday: 15
thursday: 3
friday: 2
saturday: 1
sunday: 1

Bob:
ooo_days:
monday: true
tuesday: true
wednesday: false
thursday: false
friday: false
saturday: true
sunday: false
day_prefs:
monday: 11
tuesday: 2
wednesday: 9
thursday: 6
friday: 5
saturday: 0
sunday: 1

Outcome Summary: Meeting agreed on Friday.

Final valuation for Alice: 2.
Calculation: Friday selected. Not an out of office day. Value of friday: 2.

Example 3:
{m}

Final valuation for {p}: 
'''

PAYOFF_OBS_TRANS_PREFIX = ['Read the following dialogue and extract out the ' +
                           'message that captures the final agreement made ' +
                           'between the two parties. If the players could ' +
                           'not agree, say no agreement was ' +
                           'reached. If both players agreed, say ' +
                           'which day the players agreed to meet.']
PAYOFF_OBS_TRANS_PREFIX = ('\n\n'.join(text.wrap(PAYOFF_OBS_TRANS_PREFIX)) +
                           '\n\n')

PAYOFF_OBS_TRANS_POSTFIX = ''

PAYOFF = payoff_utils.Payoff(PAYOFF_PROMPT,
                             MIN_PAYOFF,
                             MAX_PAYOFF,
                             PAYOFF_OBS_TRANS_PREFIX,
                             PAYOFF_OBS_TRANS_POSTFIX)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the game-specific functions for chess."""


from absl import flags
from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

import pyspiel
from open_spiel.python.utils import file_utils

chess = pyspiel.chess


FLAGS = flags.FLAGS

# From CMakeLists.txt:Python tests are run from the main binary directory which
# will be something like build/python.
flags.DEFINE_string(
    "chess960_fens_file",
    "../../open_spiel/games/chess/chess960_starting_positions.txt",
    "FENs database for chess960",
)


class GamesChessTest(parameterized.TestCase):

  def test_bindings_sim(self):
    game = pyspiel.load_game("chess")
    state = game.new_initial_state()
    board = None
    while not state.is_terminal():
      print(state)
      player = state.current_player()
      legal_actions = state.legal_actions()
      board = state.board()
      for action in legal_actions:
        action_str = state.action_to_string(player, action)
        move = chess.action_to_move(action, board)
        move_from = move.from_square
        move_to = move.to_square
        decoded_from_to = (f"({move_from.x} {move_from.y}) -> " +
                           f"({move_to.x} {move_to.y})")
        print(f"Legal action: {action_str} decoded from to {decoded_from_to}")
        print(f"Move representations: {move.to_string()} | " +
              f"{move.to_lan()} | {move.to_san(board)}")
        # Now do the reverse mapping from both string representations to check
        # that they correspond to this action.
        action_from_lan = state.parse_move_to_action(move.to_lan())
        action_from_san = state.parse_move_to_action(move.to_san(board))
        self.assertEqual(action, action_from_lan)
        self.assertEqual(action, action_from_san)
      action = np.random.choice(legal_actions)
      state.apply_action(action)
    print(board.to_unicode_string())
    print(board.debug_string())
    print("Moves history:")
    print(" ".join([move.to_lan() for move in state.moves_history()]))
    self.assertTrue(state.is_terminal())

  def test_state_from_fen(self):
    game = pyspiel.load_game("chess")
    fen_string = "8/k1P5/8/1K6/8/8/8/8 w - - 0 1"
    state = game.new_initial_state(fen_string)
    self.assertEqual(state.board().to_fen(), fen_string)
    self.assertEqual(state.num_repetitions(state), 1)

  @parameterized.parameters(
      "bbqnnrkr/pppppppp/8/8/8/8/PPPPPPPP/BBQNNRKR w KQkq - 0 1",
      "rnbnkbqr/pppppppp/8/8/8/8/PPPPPPPP/RNBNKBQR w KQkq - 0 1",
      "rkrnnqbb/pppppppp/8/8/8/8/PPPPPPPP/RKRNNQBB w KQkq - 0 1",
  )
  def test_chess960_sim_specific_fens(self, initial_fen):
    game = pyspiel.load_game("chess(chess960=true)")
    state = game.new_initial_state(initial_fen)
    while not state.is_terminal():
      assert not state.is_chance_node()
      legal_actions = state.legal_actions()
      action = np.random.choice(legal_actions)
      state.apply_action(action)

  def test_chess_action_conversions(self):
    game = pyspiel.load_game("chess")
    state = game.new_initial_state()
    for _ in range(10):
      while not state.is_terminal():
        assert not state.is_chance_node()
        legal_actions = state.legal_actions()
        for action in legal_actions:
          move = chess.action_to_move(action, state.board())
          move_uci = move.to_lan()
          action_mapped = chess.move_to_action(move, 8)
          self.assertEqual(
              action, action_mapped, f"Error for action {move_uci}"
          )
        action = np.random.choice(legal_actions)
        state.apply_action(action)

  def test_chess960_game_sim(self):
    fens_filename = file_utils.find_file(FLAGS.chess960_fens_file, 1)
    if fens_filename is not None:
      print("Found chess960 fens file. Running simulation tests.")
      game = pyspiel.load_game(
          f"chess(chess960=true,chess960_fens_file={fens_filename})"
      )
      for _ in range(10):
        state = game.new_initial_state()
        assert state.is_chance_node()
        outcomes = state.chance_outcomes()
        assert len(outcomes) == 960
        action_list, prob_list = zip(*outcomes)
        action = np.random.choice(action_list, p=prob_list)
        state.apply_action(action)
        while not state.is_terminal():
          assert not state.is_chance_node()
          legal_actions = state.legal_actions()
          for action in legal_actions:
            move = chess.action_to_move(action, state.board())
            move_uci = move.to_lan()
            action_mapped = chess.move_to_action(move, 8)
            self.assertEqual(
                action, action_mapped, f"Error for action {move_uci}"
            )
          action = np.random.choice(legal_actions)
          state.apply_action(action)


if __name__ == "__main__":
  np.random.seed(87375711)
  absltest.main()



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.jax.jax_cfr.

All of them are taken from open_spiel.python.algorithms.cfr_test.py
"""

import itertools

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from open_spiel.python import policy
from open_spiel.python.algorithms import expected_game_score
from open_spiel.python.jax.cfr.jax_cfr import JaxCFR
import pyspiel


class CFRTest(parameterized.TestCase, absltest.TestCase):

  def test_cfr_kuhn_poker(self):
    game = pyspiel.load_game("kuhn_poker")
    cfr_solver = JaxCFR(game)
    cfr_solver.multiple_steps(300)
    average_policy = cfr_solver.average_policy()
    average_policy_values = expected_game_score.policy_value(
        game.new_initial_state(), [average_policy] * 2
    )
    # 1/18 is the Nash value. See https://en.wikipedia.org/wiki/Kuhn_poker
    np.testing.assert_allclose(
        average_policy_values, [-1 / 18, 1 / 18], atol=1e-3
    )

  def test_cfr_plus_kuhn_poker(self):
    game = pyspiel.load_game("kuhn_poker")
    cfr_solver = JaxCFR(game)
    cfr_solver.multiple_steps(200)
    average_policy = cfr_solver.average_policy()
    average_policy_values = expected_game_score.policy_value(
        game.new_initial_state(), [average_policy] * 2
    )
    # 1/18 is the Nash value. See https://en.wikipedia.org/wiki/Kuhn_poker
    np.testing.assert_allclose(
        average_policy_values, [-1 / 18, 1 / 18], atol=1e-3
    )

  def test_cfr_plus_solver_best_response_mdp(self):
    game = pyspiel.load_game("kuhn_poker")
    cfr_solver = JaxCFR(game, True, True, True)
    cfr_solver.multiple_steps(200)
    average_policy = cfr_solver.average_policy()
    pyspiel_avg_policy = policy.python_policy_to_pyspiel_policy(average_policy)
    br_computer = pyspiel.TabularBestResponseMDP(game, pyspiel_avg_policy)
    br_info = br_computer.exploitability()
    self.assertLessEqual(br_info.exploitability, 0.001)

  @parameterized.parameters(
      list(itertools.product([True, False], [True, False], [True, False]))
  )
  def test_cfr_kuhn_poker_runs_with_multiple_players(
      self, linear_averaging, alternating_updates, regret_matching_plus
  ):
    num_players = 3

    game = pyspiel.load_game("kuhn_poker", {"players": num_players})
    cfr_solver = JaxCFR(
        game,
        regret_matching_plus=regret_matching_plus,
        alternating_updates=alternating_updates,
        linear_averaging=linear_averaging,
    )
    # for _ in range(10):
    cfr_solver.multiple_steps(10)
    average_policy = cfr_solver.average_policy()
    average_policy_values = expected_game_score.policy_value(
        game.new_initial_state(), [average_policy] * num_players
    )
    del average_policy_values


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Creates a chat game as an OpenSpiel Environment."""

from typing import Any, Callable, Dict, OrderedDict, List, Tuple, Union
from absl import logging
import numpy as np

from open_spiel.python.games.chat_games import chat_game_base
from open_spiel.python.games.chat_games.configs import config_fixed_mock
from open_spiel.python.games.chat_games.configs import config_rnd_mock
from open_spiel.python.games.chat_games.envs.observations import utils as observation_utils
from open_spiel.python.games.chat_games.envs.payoffs import utils as payoff_utils
from open_spiel.python.games.chat_games.envs.termination import utils as term_utils
from open_spiel.python.games.chat_games.envs.utils import header as header_utils
from open_spiel.python.games.chat_games.utils import test_utils as chat_test_utils

import pyspiel


GAME_TYPE = pyspiel.GameType(
        short_name='chat_game',
        long_name='Chat Game',
        utility=pyspiel.GameType.Utility.GENERAL_SUM,
        provides_information_state_string=True,
        provides_information_state_tensor=False,
        **chat_game_base.GAME_TYPE_KWARGS)


class ChatGameObserver(chat_game_base.ChatGameObserverBase):
  """Observer, conforming to the PyObserver interface (see observation.py)."""

  def _build_str_to_info_state(self) -> bool:
    """Initializes map from str to infostate. Returns True if successful."""
    # Build a string tokenizer here
    # --------------------------- #
    # Build a string tokenizer here
    return True

  def _info_state(self, input_text: str, obs_size: int) -> np.ndarray:
    """Returns a len-obs_size np.ndarray given an input string and obs_size."""
    if not self._str_to_info_state_built:
      raise ValueError('String to info state mapping not built!')
    del input_text
    # Vectorize a str (ideally lossless for info state) using a tokenizer here
    # ---------------------------------------------------------------------- #
    # Vectorize a str (ideally lossless for info state) using a tokenizer here
    return np.zeros(obs_size, dtype=np.int32)


class ChatGame(chat_game_base.BaseChatGame):
  """Chat game."""

  # pylint:disable=dangerous-default-value
  def __init__(
      self,
      params: Dict[str, Any] = chat_game_base.DEFAULT_PARAMS,
  ):
    """Constructor.

    Args:
      params: dict, parameter dict with the following keys

        num_distinct_actions- int, # of actions at each info set
        num_llm_seeds- int, # of seeds to use for generating LLM response
        num_players- int, # of speakers (action: recipient) on the message chain
        min_utility- float, minimum utility any player can attain
        max_utility- float, maximum utility any player can attain
        num_max_replies- int, total # of messages each player can send in an
          episode
    """
    self._game_loaded = False

    super().__init__(params)  # initializes self.game_info via base init
    super(chat_game_base.BaseChatGame, self).__init__(
        GAME_TYPE, self.game_info, params or dict())

  def load_chat_game(self,
                     llm_type: chat_test_utils.TestLLM,
                     observations: List[observation_utils.Observation],
                     vectorize: ...,
                     header: header_utils.Header,
                     payoffs: List[payoff_utils.Payoff],
                     aggregate_payoffs: Callable[[List[int]], float] = np.mean,
                     given_names: Union[List[str], None] = None,
                     given_llm_seeds: Union[List[int], None] = None,
                     given_prompt_actions: Union[OrderedDict[str, List[str]],
                                                 None] = None,
                     given_private_info: Union[OrderedDict[str, List[str]],
                                               None] = None,
                     initial_scenario: Union[Any, None] = None,
                     num_names: int = 2,
                     num_prompt_actions: Tuple[int, ...] = (4,),
                     num_private_info: Tuple[int, ...] = (4,),
                     examples_names: Union[List[str], None] = None,
                     examples_prompt_actions: Union[OrderedDict[str, List[str]],
                                                    None] = None,
                     examples_private_info: Union[OrderedDict[str, List[str]],
                                                  None] = None,
                     examples_scenarios: Union[List[Any], None] = None,
                     llm_list_suffix: str = 'Continue the list from here.',
                     llm_termination_prompt: Union[term_utils.Termination,
                                                   None] = None,
                     seed: Union[int, None] = None
                     ):
    """Constructor.

    Args:
      llm_type: item of enum type chat_test_utils.TestLLM
      observations: List of Observation items used for prompting llms to extract
        observations (string features) from dialogues
      vectorize: converts any length string into a length obs_size vector

      header: List of Header items used for prompting llms to take actions
        (construct messages) based on latent action variables and private
        information

      payoffs: list of Payoff items used for constructing queries and scoring
        dialogue for each agent
      aggregate_payoffs: function that maps from vector to nonnegative scalar

      given_names: list of strings representing names of players
      given_llm_seeds: list of ints to seed llm with to generate each message
      given_prompt_actions: ordered dict mapping action_keys
        (see envs/utils/header) to list of strings representing the set of
        available prompt actions (e.g., personalities or msg tones). Overrides
        examples_prompt_actions.
      given_private_info: ordered dict mapping info_keys
        (see envs/utils/header) to length-[num_players] list of strings
        representing the private information available to each player (e.g.,
        inventory / valuations of fruits). Overrides examples_private_info.
      initial_scenario: Scenario items representing an initial message

      num_names: int, # of names to generate (can be greater than # of players)
      num_prompt_actions: tuple of int, # of prompts to consider for each
        action_key (i.e., size of action space for each prompt action)
      num_private_info: tuple of int, # of private info states to consider for
        each info_key

      examples_names: list of strings representing examples of names of players
      examples_prompt_actions: ordered dict mapping action_keys
        (see envs/utils/header) to list of strings representing examples of
        prompt actions (e.g., personalities or msg tones).
      examples_private_info: ordered dict mapping info_keys
        (see envs/utils/header) to list of strings representing examples of
        private information available to players (e.g., inventory / valuations
        of fruits). Overrides examples_private_info.
      examples_scenarios: list of Scenario items used for meta-generating new
        scenarios

      llm_list_suffix: str, gets appended to a prompt to induce an llm to
        generate a list of items (different llms like different prompts).
        chinchilla likes ``, llmit likes `Continue the list from here.`
      llm_termination_prompt: Termination item w/ [attrs query,
        obs_trans_postfix, postfix]. llm will be asked to score a binary
        response `yes`/`no` given query.format(msg=last_msg) to determine
        whether the episode has reached a terminal state (e.g., deal has been
        agreed upon). default is empty string in which case llm terminal
        condition is left unused and episode terminates after
        num_players * num_max_replies

      seed: int, master seed for experiment (used to generate all subsequent
        seeds for any random generation)
    """

    # Define LLM model here
    self._llm_type = llm_type
    if self._llm_type == chat_test_utils.TestLLM.MOCK:
      self._lm = chat_test_utils.MockLLM()
    else:
      raise NotImplementedError(f'llm_type {self._llm_type} not available.')
    # Define LLM model here

    super()._load_chat_game(observations,
                            vectorize,
                            header,
                            payoffs,
                            aggregate_payoffs,
                            given_names,
                            given_llm_seeds,
                            given_prompt_actions,
                            given_private_info,
                            initial_scenario,
                            num_names,
                            num_prompt_actions,
                            num_private_info,
                            examples_names,
                            examples_prompt_actions,
                            examples_private_info,
                            examples_scenarios,
                            llm_list_suffix,
                            llm_termination_prompt,
                            seed)

    self._game_loaded = True

  def generate_response(self, prompt: str, seed: int,
                        num_output_tokens: Union[int, None] = None) -> str:
    """Returns LLM generated string given prompt and seed."""
    # Define generate response here
    if self._llm_type == chat_test_utils.TestLLM.MOCK:
      return self._lm.generate_response(prompt, seed, num_output_tokens)
    else:
      raise NotImplementedError(f'llm_type {self._llm_type} not available.')
    # Define generate response here

  def generate_bool(self, prompt: str, seed: int) -> bool:
    """Returns LLM generated boolean given prompt and seed."""
    # Define generate bool here (e.g., for terminating an episode)
    if self._llm_type == chat_test_utils.TestLLM.MOCK:
      return self._lm.generate_bool(prompt, seed)
    else:
      raise NotImplementedError(f'llm_type {self._llm_type} not available.')
    # Define generate bool here

  def make_py_observer(self,
                       iig_obs_type: Union[pyspiel.IIGObservationType,
                                           None] = None,
                       params: Union[Dict[str, Any], None] = None
                       ) -> ChatGameObserver:
    """Returns an object used for observing game state."""
    return ChatGameObserver(
        iig_obs_type or pyspiel.IIGObservationType(perfect_recall=False),
        params)

  def new_initial_state(self) -> chat_game_base.ChatGameState:
    """Generates a new dialogue game.

    Returns:
      chat_game_base.ChatGameState (see chat_games/chat_game_base.py)
    """
    # KEEP THIS IF-BLOCK FOR OPEN_SPIEL TESTS
    if not self._game_loaded:
      # load mock game for testing
      if self._num_players == 2:
        config = config_fixed_mock.get_config()
        tones = config.game.given_prompt_actions.values()[0]
        num_prompt_actions = (len(tones),)
      else:
        config = config_rnd_mock.get_config()
        num_prompt_actions = config.game.num_prompt_actions
      # open_spiel attempts to run several simulation tests of games. this
      # chat_game, however, requires calling `load_chat_game` explicitly after
      # __init__ which is unique. we do this because the most obvious place to
      # pass game configs would be via `params`, but everything in params must
      # be `pickleable` which rules out passing things like `vectorizers` and
      # messsy llm string generators. therefore, we need to check to see if
      # `load_chat_game` has been called here and call it if not.
      # also, open_spiel tests run with variable numbers of players which are
      # different from those in chat_game_base.DEFAULT_PARAMS. More importantly,
      # this affects the number of distinct actions since the number of players
      # affects who we can choose to speak to. hence, we explicitly recalculate
      # the number of distinct actions here (overwriting what was specified in
      # the original chat_game_base.DEFAULT_PARAMS)
      self._num_distinct_actions = np.prod(num_prompt_actions +
                                           (self._num_players,))
      vectorizer = chat_test_utils.MockVectorizer()
      self.load_chat_game(llm_type=chat_test_utils.TestLLM.MOCK,
                          vectorize=vectorizer.vectorize,
                          seed=1234,
                          **config.game)
      logging.warning('Loading chat_game with default config. Only meant for ' +
                      'open_spiel testing.')

    return chat_game_base.ChatGameState(self, self._initial_state_configs)

# Register the game with the OpenSpiel library

pyspiel.register_game(GAME_TYPE, ChatGame)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A base environment for trading fruit with private info.
"""

import dataclasses

from open_spiel.python.games.chat_games.envs.comm_substrates import trades
from open_spiel.python.games.chat_games.envs.scenarios.domains import trade_fruit
from open_spiel.python.games.chat_games.envs.utils import header
from open_spiel.python.games.chat_games.envs.utils import text


action_keys = tuple([])
info_keys = tuple(['fruit_endowment', 'fruit_valuations'])

w_opts = (trades.W_OPTS_PREFIX +
          'Fruit Endowment:\n{fruit_endowment}\n\n' +
          'Fruit Valuations:\n{fruit_valuations}\n' +
          trades.PLAIN)

# Example a
email_1a = ['Hi Joel,',
            'I would like to trade you 2 strawberries for 3 blueberries.',
            'Would you like to trade with me?',
            'Best,', 'Bob']
email_1a = (trades.PLAIN.format(sender='Alicia', receiver='Joel') +
            '\n\n'.join(text.wrap(email_1a)))

email_2a = ['Hi Alicia,',
            'Thanks for reaching out. I only have 2 blueberries, but even if ' +
            'I had 3, I would not want to give them up. Also, I dislike ' +
            'strawberries. I do not think a trade makes sense in this case.',
            'Thanks for considering trading with me though!',
            'Best,', 'Joel']
email_2a = (trades.PLAIN.format(sender='Joel', receiver='Alicia') +
            '\n\n'.join(text.wrap(email_2a)))

email_3a = ['Hi Joel,',
            'That is all well. I understand.',
            'Have a good day!',
            'Best,', 'Alicia']
email_3a = (trades.PLAIN.format(sender='Alicia', receiver='Joel') +
            '\n\n'.join(text.wrap(email_3a)))

example_a = email_1a + email_2a
example_a = example_a.strip('\n')

# Example b
email_1b = ['Hi Marcus,',
            'I would like to trade you 2 kiwis for 1 watermelon.',
            'Would you like to trade with me?',
            'Best,', 'Taylor']
email_1b = (trades.PLAIN.format(sender='Taylor', receiver='Marcus') +
            '\n\n'.join(text.wrap(email_1b)))

email_2b = ['Hi Taylor,',
            'I love kiwis! And lucky for you, I have a watermelon.',
            'Lets trade!',
            'Best,', 'Marcus']
email_2b = (trades.PLAIN.format(sender='Marcus', receiver='Taylor') +
            '\n\n'.join(text.wrap(email_2b)))

email_3b = ['Hi Marcus,',
            'Great! It was a pleasure negotiating with you.',
            'Have a good day!',
            'Best,', 'Taylor']
email_3b = (trades.PLAIN.format(sender='Taylor', receiver='Marcus') +
            '\n\n'.join(text.wrap(email_3b)))

example_b = email_1b + email_2b + email_3b
example_b = example_b.strip('\n')

# Example c
email_1c = ['Hi Suzy,',
            'I would like to trade you 1 banana for 1 apple.',
            'Would you like to trade with me?',
            'Best,', 'Bob']
email_1c = (trades.PLAIN.format(sender='Bob', receiver='Suzy') +
            '\n\n'.join(text.wrap(email_1c)))

email_2c = ['Hi Bob,',
            'Thanks for reaching out. I really like my apples so I am ' +
            'hesitant to give them up. Would you be willing to take a few ' +
            'kiwis instead? I would like to trade you 3 kiwis for 1 banana.',
            'Does that work?',
            'Best,', 'Suzy']
email_2c = (trades.PLAIN.format(sender='Suzy', receiver='Bob') +
            '\n\n'.join(text.wrap(email_2c)))

email_3c = ['Hi Suzy,',
            'Yes! I would have preferred an apple but 3 kiwis are nearly as ' +
            'good and I would rather have those than a banana.',
            'Thanks for trading with me!',
            'Best,', 'Bob']
email_3c = '\n\n'.join(text.wrap(email_3c))

example_c = email_1c + email_2c
example_c = example_c.strip('\n')

instr_a = ['You are an assistant who is playing a game where you trade fruit.' +
           ' You want to make a trade that is best for you. You will read a ' +
           'dialogue that contains a conversation where you have been ' +
           'negotiating to trade your fruit for another persons fruit. You ' +
           'will then read a text block that contains information a) about ' +
           'the actual fruit you currently have and are able to trade and b)' +
           ' information about how much you value certain types of fruit.',
           'You should use everything you learned from this to decide to ',
           '1) accept the trade if you are happy with the terms,',
           '2) reject the negotiation all together and say goodbye if you do ' +
           'not think an agreement can be reached,',
           '3) counter-propose an alternative trade that includes what fruit ' +
           'you would like to give and what fruit you would like to receive ' +
           'in turn.',
           'Consider the following example dialogues. Components of the ' +
           'examples will be demarked with the symbol "&". Here is the first ' +
           'example which shows a trade is rejected.',
           '&' * 50]
instr_b = ['&' * 50,
           'Here is a second example where a trade is accepted.',
           '&' * 50]
instr_c = ['&' * 50,
           'Here is a partial dialogue where we demonstrate a reasonable ' +
           'countertrade.',
           '&' * 50]
instr_d = ['&' * 50,
           'Continuing the example. You now see the fruit you have and how ' +
           'much you value each fruit type.',
           '&' * 50]
info = w_opts.format(sender='Bob', receiver='Suzy',
                     fruit_endowment=trade_fruit.ENDOWMENT_A,
                     fruit_valuations=trade_fruit.VALUATION_A).strip('\n')
instr_e = ['&' * 50,
           'A reasonable way to respond would be as follows:',
           '&' * 50]
instr_f = ['&' * 50,
           'Now you are going to read a fresh dialogue, fruit endowment, and ' +
           'fruit valuation information. Please give a reasonable response ' +
           'that attempts to reach an agreement to trade fruit.',
           '&' * 50]
context = (text.wrap(instr_a) + [example_a] + text.wrap(instr_b) +[example_b] +
           text.wrap(instr_c) + [example_c] + text.wrap(instr_d) + [info] +
           text.wrap(instr_e) + [email_3c] + text.wrap(instr_f))

HEADER = header.Header(trades.PLAIN,
                       w_opts,
                       trades.strip_msg,
                       trades.SPECIAL_CHARS,
                       action_keys,
                       info_keys,
                       '\n\n'.join(context))


@dataclasses.dataclass(frozen=True)
class Scenario(header.BaseScenario):
  fruit_endowment: str
  fruit_valuations: str



# coding=utf-8
# Copyright 2018 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A checkpointing mechanism for Dopamine agents.

This Checkpointer expects a base directory where checkpoints for different
iterations are stored. Specifically, Checkpointer.save_checkpoint() takes in
as input a dictionary 'data' to be pickled to disk. At each iteration, we
write a file called 'cpkt.#', where # is the iteration number. The
Checkpointer also cleans up old files, maintaining up to the CHECKPOINT_DURATION
most recent iterations.

The Checkpointer writes a sentinel file to indicate that checkpointing was
globally successful. This means that all other checkpointing activities
(saving the Tensorflow graph, the replay buffer) should be performed *prior*
to calling Checkpointer.save_checkpoint(). This allows the Checkpointer to
detect incomplete checkpoints.

#### Example

After running 10 iterations (numbered 0...9) with base_directory='/checkpoint',
the following files will exist:
```
  /checkpoint/cpkt.6
  /checkpoint/cpkt.7
  /checkpoint/cpkt.8
  /checkpoint/cpkt.9
  /checkpoint/sentinel_checkpoint_complete.6
  /checkpoint/sentinel_checkpoint_complete.7
  /checkpoint/sentinel_checkpoint_complete.8
  /checkpoint/sentinel_checkpoint_complete.9
```
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import pickle
import tensorflow as tf

CHECKPOINT_DURATION = 4


def get_latest_checkpoint_number(base_directory):
  """Returns the version number of the latest completed checkpoint.

  Args:
    base_directory: str, directory in which to look for checkpoint files.

  Returns:
    int, the iteration number of the latest checkpoint, or -1 if none was found.
  """
  glob = os.path.join(base_directory, 'sentinel_checkpoint_complete.*')
  def extract_iteration(x):
    return int(x[x.rfind('.') + 1:])
  try:
    checkpoint_files = tf.gfile.Glob(glob)
  except tf.errors.NotFoundError:
    return -1
  try:
    latest_iteration = max(extract_iteration(x) for x in checkpoint_files)
    return latest_iteration
  except ValueError:
    return -1


class Checkpointer(object):
  """Class for managing checkpoints for Dopamine agents.
  """

  def __init__(self, base_directory, checkpoint_file_prefix='ckpt',
               checkpoint_frequency=1):
    """Initializes Checkpointer.

    Args:
      base_directory: str, directory where all checkpoints are saved/loaded.
      checkpoint_file_prefix: str, prefix to use for naming checkpoint files.
      checkpoint_frequency: int, the frequency at which to checkpoint.

    Raises:
      ValueError: if base_directory is empty, or not creatable.
    """
    if not base_directory:
      raise ValueError('No path provided to Checkpointer.')
    self._checkpoint_file_prefix = checkpoint_file_prefix
    self._checkpoint_frequency = checkpoint_frequency
    self._base_directory = base_directory
    try:
      tf.gfile.MakeDirs(base_directory)
    except tf.errors.PermissionDeniedError:
      # We catch the PermissionDeniedError and issue a more useful exception.
      raise ValueError('Unable to create checkpoint path: {}.'.format(
          base_directory))

  def _generate_filename(self, file_prefix, iteration_number):
    """Returns a checkpoint filename from prefix and iteration number."""
    filename = '{}.{}'.format(file_prefix, iteration_number)
    return os.path.join(self._base_directory, filename)

  def _save_data_to_file(self, data, filename):
    """Saves the given 'data' object to a file."""
    with tf.gfile.GFile(filename, 'w') as fout:
      pickle.dump(data, fout)

  def save_checkpoint(self, iteration_number, data):
    """Saves a new checkpoint at the current iteration_number.

    Args:
      iteration_number: int, the current iteration number for this checkpoint.
      data: Any (picklable) python object containing the data to store in the
        checkpoint.
    """
    if iteration_number % self._checkpoint_frequency != 0:
      return

    filename = self._generate_filename(self._checkpoint_file_prefix,
                                       iteration_number)
    self._save_data_to_file(data, filename)
    filename = self._generate_filename('sentinel_checkpoint_complete',
                                       iteration_number)
    with tf.gfile.GFile(filename, 'wb') as fout:
      fout.write('done')

    self._clean_up_old_checkpoints(iteration_number)

  def _clean_up_old_checkpoints(self, iteration_number):
    """Removes sufficiently old checkpoints."""
    # After writing a the checkpoint and sentinel file, we garbage collect files
    # that are CHECKPOINT_DURATION * self._checkpoint_frequency versions old.
    stale_iteration_number = iteration_number - (self._checkpoint_frequency *
                                                 CHECKPOINT_DURATION)

    if stale_iteration_number >= 0:
      stale_file = self._generate_filename(self._checkpoint_file_prefix,
                                           stale_iteration_number)
      stale_sentinel = self._generate_filename('sentinel_checkpoint_complete',
                                               stale_iteration_number)
      try:
        tf.gfile.Remove(stale_file)
        tf.gfile.Remove(stale_sentinel)
      except tf.errors.NotFoundError:
        # Ignore if file not found.
        tf.logging.info('Unable to remove {} or {}.'.format(stale_file,
                                                            stale_sentinel))

  def _load_data_from_file(self, filename):
    if not tf.gfile.Exists(filename):
      return None
    with tf.gfile.GFile(filename, 'rb') as fin:
      return pickle.load(fin)

  def load_checkpoint(self, iteration_number):
    """Tries to reload a checkpoint at the selected iteration number.

    Args:
      iteration_number: The checkpoint iteration number to try to load.

    Returns:
      If the checkpoint files exist, two unpickled objects that were passed in
        as data to save_checkpoint; returns None if the files do not exist.
    """
    checkpoint_file = self._generate_filename(self._checkpoint_file_prefix,
                                              iteration_number)
    return self._load_data_from_file(checkpoint_file)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test Python bindings for game transforms."""

from absl.testing import absltest

import numpy as np

from open_spiel.python.algorithms import cfr
from open_spiel.python.algorithms import expected_game_score
from open_spiel.python.games import iterated_prisoners_dilemma as ipd  # pylint: disable=unused-import
import pyspiel


SEED = 1098097


class GameTransformsTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(SEED)

  def test_create_repeated_game(self):
    """Test both create_repeated_game function signatures."""
    repeated_game = pyspiel.create_repeated_game("matrix_rps",
                                                 {"num_repetitions": 10})
    assert repeated_game.utility_sum() == 0
    state = repeated_game.new_initial_state()
    for _ in range(10):
      state.apply_actions([0, 0])
    assert state.is_terminal()

    stage_game = pyspiel.load_game("matrix_mp")
    repeated_game = pyspiel.create_repeated_game(stage_game,
                                                 {"num_repetitions": 5})
    state = repeated_game.new_initial_state()
    for _ in range(5):
      state.apply_actions([0, 0])
    assert state.is_terminal()

    stage_game = pyspiel.load_game("matrix_pd")
    repeated_game = pyspiel.create_repeated_game(stage_game,
                                                 {"num_repetitions": 5})
    assert repeated_game.utility_sum() is None

  def test_cached_tree_sim(self):
    """Test both create_cached_tree function signatures."""
    for game_name in ["kuhn_poker", "python_tic_tac_toe"]:
      cached_tree_game = pyspiel.convert_to_cached_tree(
          pyspiel.load_game(game_name))
      assert cached_tree_game.num_players() == 2
      for _ in range(10):
        state = cached_tree_game.new_initial_state()
        while not state.is_terminal():
          legal_actions = state.legal_actions()
          action = np.random.choice(legal_actions)
          state.apply_action(action)
        self.assertTrue(state.is_terminal())

  def test_cached_tree_cfr_kuhn(self):
    game = pyspiel.load_game("cached_tree(game=kuhn_poker())")
    cfr_solver = cfr.CFRSolver(game)
    for _ in range(300):
      cfr_solver.evaluate_and_update_policy()
    average_policy = cfr_solver.average_policy()
    average_policy_values = expected_game_score.policy_value(
        game.new_initial_state(), [average_policy] * 2)
    # 1/18 is the Nash value. See https://en.wikipedia.org/wiki/Kuhn_poker
    np.testing.assert_allclose(
        average_policy_values, [-1 / 18, 1 / 18], atol=1e-3)

  def test_turn_based_simultaneous_game(self):
    tb_game = pyspiel.load_game(
        "turn_based_simultaneous_game(game="
        "goofspiel(num_cards=13,players=2,points_order=descending))")
    state = tb_game.new_initial_state()
    # Play 11 moves. This gets to the decision right before the game is over.
    # Because the last action is taken automatically there are only 12 moves
    # total.
    for _ in range(11):
      state.apply_action(state.legal_actions()[0])
      state.apply_action(state.legal_actions()[0])
    sim_state = state.simultaneous_game_state()
    assert not state.is_terminal()
    assert not sim_state.is_terminal()
    # For the last joint action, pull out the simultaneous state from inside the
    # wrapper and and apply joint action to it. Both the wrapped state and
    # the simultaneous state should be terminal after this.
    sim_state.apply_actions([sim_state.legal_actions(0)[0],
                             sim_state.legal_actions(1)[0]])
    assert state.is_terminal()
    assert sim_state.is_terminal()

  def test_turn_based_simultaneous_python_game(self):
    tb_game = pyspiel.load_game(
        "turn_based_simultaneous_game(game="
        "python_iterated_prisoners_dilemma())"
    )
    state = tb_game.new_initial_state()
    # Play 10 rounds, then continue.
    for _ in range(10):
      state.apply_action(state.legal_actions()[0])
      state.apply_action(state.legal_actions()[0])
      if state.is_chance_node():
        state.apply_action(ipd.Chance.CONTINUE)
    # Pull out the simultaneous state from inside the wrapper.
    sim_state = state.simultaneous_game_state()
    assert not state.is_terminal()
    assert not sim_state.is_terminal()
    sim_state.apply_actions([sim_state.legal_actions(0)[0],
                             sim_state.legal_actions(1)[0]])
    assert not state.is_terminal()
    assert not sim_state.is_terminal()
    # Cannot properly check is_chance_node() because the wrapper is still in
    # rollout mode, so this would fail: assert state.is_chance_node()
    assert sim_state.is_chance_node()
    sim_state.apply_action(ipd.Chance.STOP)
    assert state.is_terminal()
    assert sim_state.is_terminal()


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Treatment of iterates and gradients over the simplex."""

from absl import logging  # pylint:disable=unused-import

import numpy as np


def grad_norm(dist, grad, eps=1e-8, simplex_tol=1e-9):
  """Compute norm of gradient projected onto the tangent space of simplex.

  *assumes context is gradient descent (not ascent)

  Args:
    dist: np.array, distribution
    grad: np.array, gradient (same shape as distribution)
    eps: float, elements of dist in [eps, 1 - eps] are considered to be in the
      interior of the simplex. gradients on border of simplex
    simplex_tol: float, tolerance for checking if a point lies on the simplex,
      sum(vec) <= 1 + simplex_tol and all(vec > -simplex_tol). should be smaller
      than eps descent steps or points that are "leaving" simplex will be
      mislabeled
  Returns:
    float, norm of projected gradient
  """
  if simplex_tol >= eps:
    raise ValueError("simplex_tol should be less than eps")
  grad_proj = project_grad(grad)
  g_norm = np.linalg.norm(grad_proj)
  if g_norm > 0:
    # take a gradient descent step in the direction grad_proj with len eps
    # to determine if the update is "leaving" the simplex
    dist -= eps * grad_proj / g_norm
    if not ((np.sum(dist) <= 1 + simplex_tol) and np.all(dist >= -simplex_tol)):
      g_norm = 0.
  return g_norm


def project_grad(g):
  """Project gradient onto tangent space of simplex."""
  return g - g.sum() / g.size


# Project to probability simplex
# Based on this paper:
# Projection onto the probability simplex: An efficient algorithm with a
# simple proof, and an application
# https://arxiv.org/pdf/1309.1541.pdf
def euclidean_projection_onto_simplex(y, eps=1e-3, subset=True):
  """O(n log n) Euclidean projection of y onto the simplex.

  Args:
    y: np.array
    eps: float, ensure x remains at least eps / dim away from facets of simplex
    subset: bool, whether to project onto a subset of the simplex defined by eps
  Returns:
    np.array, y projected onto the simplex
  """
  if np.all(y >= 0.) and np.abs(np.sum(y) - 1.) < 1e-8:
    return y
  d = len(y)
  u = sorted(y, reverse=True)
  sum_uj = 0.
  rho = 0.
  for j in range(d):
    sum_uj += u[j]
    tj = (1. - sum_uj) / (j + 1.)
    if u[j] + tj <= 0:
      rho = j - 1
      sum_uj = sum_uj - u[j]
      break
    else:
      rho = j
  lam = (1. - sum_uj) / (rho + 1.)
  x = np.array([max(y[i] + lam, 0.) for i in range(d)])
  if subset:
    scale = 1. - eps * float(d + 1) / d
    offset = eps / float(d)
    x = scale * x + offset
    x /= x.sum()
  return x


def project_to_interior(x, eps):
  """Project x onto interior of simplex.

  Args:
    x: np.array of shape (dim,)
    eps: float, ensure x remains at least eps / dim away from facets of simplex
  Returns:
    np.array, distribution x with min(x) >= eps / dim
  """
  min_x = np.min(x)
  d = len(x)
  if min_x < eps / d:
    t = (eps / d - min_x) / (1. / d - min_x)
    x = x * (1 - t) + 1 / d * t
  return x



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the game-specific functions for tic_tac_toe."""


from absl.testing import absltest

import pyspiel
ttt = pyspiel.tic_tac_toe


def make_game():
  return pyspiel.load_game("tic_tac_toe")


class GamesTicTacToeTest(absltest.TestCase):

  def test_constants(self):
    self.assertEqual(ttt.NUM_ROWS, 3)
    self.assertEqual(ttt.NUM_COLS, 3)
    self.assertEqual(ttt.NUM_CELLS, 9)
    self.assertEqual(ttt.CellState.EMPTY.value, 0)
    self.assertEqual(ttt.CellState.NOUGHT.value, 1)
    self.assertEqual(ttt.CellState.CROSS.value, 2)

  def test_player_to_cellstate(self):
    self.assertEqual(ttt.player_to_cellstate(0),
                     ttt.CellState.CROSS)
    self.assertEqual(ttt.player_to_cellstate(1),
                     ttt.CellState.NOUGHT)

  def test_cellstate_to_string(self):
    self.assertEqual(ttt.cellstate_to_string(ttt.CellState.EMPTY), ".")
    self.assertEqual(ttt.cellstate_to_string(ttt.CellState.NOUGHT), "o")
    self.assertEqual(ttt.cellstate_to_string(ttt.CellState.CROSS), "x")

  def test_board_at(self):
    game = make_game()
    state = game.new_initial_state()
    state.apply_action(4)
    self.assertEqual(state.board_at(1, 1), ttt.CellState.CROSS)

  def test_board(self):
    game = make_game()
    state = game.new_initial_state()
    state.apply_action(0)
    state.apply_action(1)
    self.assertEqual(state.board(), [
        ttt.CellState.CROSS,
        ttt.CellState.NOUGHT] + [ttt.CellState.EMPTY] * 7)

  def test_json(self):
    game = make_game()
    state = game.new_initial_state()
    state.apply_action(4)
    state_struct = state.to_struct()
    self.assertEqual(
        state_struct.board,
        [".", ".", ".", ".", "x", ".", ".", ".", "."],
    )
    self.assertEqual(state_struct.current_player, "o")
    json_from_struct = state_struct.to_json()
    state_json = state.to_json()
    self.assertEqual(
        state_json,
        '{"board":[".",".",".",".","x",".",".",".","."],"current_player":"o"}',
    )
    self.assertEqual(json_from_struct, state_json)
    state_struct = ttt.TicTacToeStateStruct(state_json)
    self.assertEqual(state_struct.to_json(), state_json)


if __name__ == "__main__":
  absltest.main()




# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TD Learning with N-Tuple Networks for 2048."""

from absl import app
from absl import flags
from absl import logging
import numpy as np
import pyspiel

FLAGS = flags.FLAGS
flags.DEFINE_string("game", "2048", "Name of the game.")
flags.DEFINE_integer("num_train_episodes", 15000,
                     "Number of training episodes.")
flags.DEFINE_integer("eval_every", 1000,
                     "Episode frequency at which the agent is evaluated.")
flags.DEFINE_float("alpha", 0.02, "Learning rate")


class NTupleNetwork:
  """An N-tuple Network class.

  N-Tuple Networks are an effective way of reducing the storage requirement for
  evaluating and learning state values. This is accomplished by defining a
  collection of N-Tuples that represent various segments in a game's
  ObservationTensor.

  The value of a given state is defined as the sum of values of each N-Tuple,
  which are stored in a look up table. The policy of the agent is to chose an
  action that maximises the value of the after-state. After each episode, all
  the states that were reached in that episode is used for updating the state
  values using Temporal Difference Learning.

  References:
  [1] Szubert, Marcin and Wojciech Jaśkowski. "Temporal difference learning of
  n-tuple networks for the game 2048." Computational Intelligence and Games
  (CIG), 2014 IEEE Conference on. IEEE, 2014.
  """

  def __init__(self, n_tuple_size, max_tuple_index, n_tuples):
    for tuples in n_tuples:
      if len(tuples) != n_tuple_size:
        raise ValueError("n_tuple_size does not match size of tuples")
    n_tuple_network_size = len(n_tuples)
    look_up_table_shape = (n_tuple_network_size,) + (
        max_tuple_index,
    ) * n_tuple_size

    self.n_tuples = n_tuples
    self.look_up_table = np.zeros(look_up_table_shape)

  def learn(self, states):
    target = 0
    while states:
      state = states.pop()
      error = target - self.value(state)
      target = state.rewards()[0] + self.update(state, FLAGS.alpha * error)

  def update(self, state, adjust):
    v = 0
    for idx, n_tuple in enumerate(self.n_tuples):
      v += self.update_tuple(idx, n_tuple, state, adjust)
    return v

  def update_tuple(self, idx, n_tuple, state, adjust):
    observation_tensor = state.observation_tensor(0)
    index = (idx,) + tuple(
        [
            0
            if observation_tensor[tile] == 0
            else int(np.log2(observation_tensor[tile]))
            for tile in n_tuple
        ]
    )
    self.look_up_table[index] += adjust
    return self.look_up_table[index]

  def evaluator(self, state, action):
    working_state = state.clone()
    working_state.apply_action(action)
    return working_state.rewards()[0] + self.value(working_state)

  def value(self, state):
    """Returns the value of this state."""

    observation_tensor = state.observation_tensor(0)
    v = 0
    for idx, n_tuple in enumerate(self.n_tuples):
      lookup_tuple_index = [
          0
          if observation_tensor[tile] == 0
          else int(np.log2(observation_tensor[tile]))
          for tile in n_tuple
      ]
      lookup_index = (idx,) + tuple(lookup_tuple_index)
      v += self.look_up_table[lookup_index]
    return v


def main(_):
  n_tuple_network = NTupleNetwork(
      6,
      15,
      [
          [0, 1, 2, 3, 4, 5],
          [4, 5, 6, 7, 8, 9],
          [0, 1, 2, 4, 5, 6],
          [4, 5, 6, 8, 9, 10],
      ],
  )
  game = pyspiel.load_game(FLAGS.game)
  sum_rewards = 0
  largest_tile = 0
  max_score = 0
  for ep in range(FLAGS.num_train_episodes):
    state = game.new_initial_state()
    states_in_episode = []
    while not state.is_terminal():
      if state.is_chance_node():
        outcomes = state.chance_outcomes()
        action_list, prob_list = zip(*outcomes)
        action = np.random.choice(action_list, p=prob_list)
        state.apply_action(action)
      else:
        legal_actions = state.legal_actions(state.current_player())
        # pylint: disable=cell-var-from-loop
        best_action = max(
            legal_actions,
            key=lambda action: n_tuple_network.evaluator(state, action),
        )
        state.apply_action(best_action)
        states_in_episode.append(state.clone())

    sum_rewards += state.returns()[0]
    largest_tile_from_episode = max(state.observation_tensor(0))
    if largest_tile_from_episode > largest_tile:
      largest_tile = largest_tile_from_episode
    if state.returns()[0] > max_score:
      max_score = state.returns()[0]

    n_tuple_network.learn(states_in_episode)

    if (ep + 1) % FLAGS.eval_every == 0:
      logging.info(
          "[%s] Average Score: %s, Max Score: %s, Largest Tile Reached: %s",
          ep + 1,
          int(sum_rewards / FLAGS.eval_every),
          int(max_score),
          int(largest_tile),
      )
      sum_rewards = 0
      largest_tile = 0
      max_score = 0


if __name__ == "__main__":
  app.run(main)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.voting.util."""

from absl.testing import absltest
from open_spiel.python.voting import preflib_util

TEST_DATA = """
# FILE NAME: 00004-00000050.soc
# TITLE: Netflix Prize Data
# DESCRIPTION:
# DATA TYPE: soc
# MODIFICATION TYPE: induced
# RELATES TO:
# RELATED FILES:
# PUBLICATION DATE: 2013-08-17
# MODIFICATION DATE: 2022-09-16
# NUMBER ALTERNATIVES: 3
# NUMBER VOTERS: 391
# NUMBER UNIQUE ORDERS: 6
# ALTERNATIVE NAME 1: The Amityville Horror
# ALTERNATIVE NAME 2: Mars Attacks!
# ALTERNATIVE NAME 3: Lean on Me
186: 3,1,2
71: 1,3,2
58: 3,2,1
45: 2,3,1
18: 1,2,3
13: 2,1,3
"""


class UtilTest(absltest.TestCase):
  def test_load_preflib(self):
    print(TEST_DATA)
    profile = preflib_util.parse_preflib_data(TEST_DATA)
    print(profile)
    self.assertEqual(profile.num_alternatives(), 3)
    self.assertEqual(profile.num_votes(), 391)
    self.assertListEqual(profile.alternatives, [
        "The Amityville Horror", "Mars Attacks!", "Lean on Me"
    ])
    print(profile.alternatives)
    print(profile.margin_matrix())


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np
import torch

from open_spiel.python import policy
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import exploitability
import pyspiel
from open_spiel.python.pytorch import policy_gradient
from open_spiel.python.pytorch.losses import rl_losses

SEED = 24984617


class PolicyGradientPolicies(policy.Policy):
  """Joint policy to be evaluated."""

  def __init__(self, env, nfsp_policies):
    game = env.game
    player_ids = [0, 1]
    super(PolicyGradientPolicies, self).__init__(game, player_ids)
    self._policies = nfsp_policies
    self._obs = {"info_state": [None, None], "legal_actions": [None, None]}

  def action_probabilities(self, state, player_id=None):
    cur_player = state.current_player()
    legal_actions = state.legal_actions(cur_player)

    self._obs["current_player"] = cur_player
    self._obs["info_state"][cur_player] = state.information_state_tensor(
        cur_player
    )
    self._obs["legal_actions"][cur_player] = legal_actions

    info_state = rl_environment.TimeStep(
        observations=self._obs, rewards=None, discounts=None, step_type=None
    )

    p = self._policies[cur_player].step(info_state, is_evaluation=True).probs
    prob_dict = {action: p[action] for action in legal_actions}
    return prob_dict


class PolicyGradientTest(parameterized.TestCase, absltest.TestCase):

  @parameterized.parameters(
      itertools.product(("rpg", "qpg", "rm", "a2c", "neurd"),
                        ("kuhn_poker", "leduc_poker")))
  def test_run_game(self, loss_str, game_name):
    env = rl_environment.Environment(game_name)
    env.seed(SEED)
    info_state_size = env.observation_spec()["info_state"][0]
    num_actions = env.action_spec()["num_actions"]

    agents = [
        policy_gradient.PolicyGradient(  # pylint: disable=g-complex-comprehension
            player_id=player_id,
            info_state_size=info_state_size,
            num_actions=num_actions,
            loss_str=loss_str,
            hidden_layers_sizes=[32, 32],
            batch_size=16,
            entropy_cost=0.001,
            critic_learning_rate=0.01,
            pi_learning_rate=0.01,
            num_critic_before_pi=4) for player_id in [0, 1]
    ]

    for _ in range(2):
      time_step = env.reset()
      while not time_step.last():
        current_player = time_step.observations["current_player"]
        current_agent = agents[current_player]
        agent_output = current_agent.step(time_step)
        time_step = env.step([agent_output.action])

      for agent in agents:
        agent.step(time_step)

  def test_neurd_kuhn(self):
    env = rl_environment.Environment("kuhn_poker")
    env.seed(SEED)
    info_state_size = env.observation_spec()["info_state"][0]
    num_actions = env.action_spec()["num_actions"]

    agents = [
        policy_gradient.PolicyGradient(  # pylint: disable=g-complex-comprehension
            player_id=player_id,
            info_state_size=info_state_size,
            num_actions=num_actions,
            loss_str="neurd",
            hidden_layers_sizes=[32],
            batch_size=16,
            entropy_cost=0.001,
            critic_learning_rate=0.01,
            pi_learning_rate=0.01,
            num_critic_before_pi=4) for player_id in [0, 1]
    ]
    expl_policies_avg = PolicyGradientPolicies(env, agents)

    for _ in range(100):
      time_step = env.reset()
      while not time_step.last():
        current_player = time_step.observations["current_player"]
        current_agent = agents[current_player]
        agent_output = current_agent.step(time_step)
        time_step = env.step([agent_output.action])

      for agent in agents:
        agent.step(time_step)

    expl = exploitability.exploitability(env.game, expl_policies_avg)
    # Check the exploitability is less than the target upper bound.
    self.assertLess(expl, 0.7)

  def test_run_hanabi(self):
    # Hanabi is an optional game, so check we have it before running the test.
    game = "hanabi"
    if game not in pyspiel.registered_names():
      return

    num_players = 3
    env_configs = {
        "players": num_players,
        "max_life_tokens": 1,
        "colors": 2,
        "ranks": 3,
        "hand_size": 2,
        "max_information_tokens": 3,
        "discount": 0.99
    }
    env = rl_environment.Environment(game, **env_configs)
    env.seed(SEED)
    info_state_size = env.observation_spec()["info_state"][0]
    num_actions = env.action_spec()["num_actions"]

    agents = [
        policy_gradient.PolicyGradient(  # pylint: disable=g-complex-comprehension
            player_id=player_id,
            info_state_size=info_state_size,
            num_actions=num_actions,
            hidden_layers_sizes=[8, 8],
            batch_size=16,
            entropy_cost=0.001,
            critic_learning_rate=0.001,
            pi_learning_rate=0.001,
            num_critic_before_pi=4) for player_id in range(num_players)
    ]

    time_step = env.reset()
    while not time_step.last():
      current_player = time_step.observations["current_player"]
      agent_output = [agent.step(time_step) for agent in agents]
      time_step = env.step([agent_output[current_player].action])

    for agent in agents:
      agent.step(time_step)

  def test_loss_modes(self):
    loss_dict = {
        "qpg": rl_losses.BatchQPGLoss,
        "rpg": rl_losses.BatchRPGLoss,
        "rm": rl_losses.BatchRMLoss,
        "a2c": rl_losses.BatchA2CLoss,
        "neurd": rl_losses.BatchNeuRDLoss,
    }

    for loss_str, loss_class in loss_dict.items():
      agent_by_str = policy_gradient.PolicyGradient(
          player_id=0,
          info_state_size=32,
          num_actions=2,
          loss_str=loss_str,
          loss_class=None)
      agent_by_class = policy_gradient.PolicyGradient(
          player_id=0,
          info_state_size=32,
          num_actions=2,
          loss_str=None,
          loss_class=loss_class)

      self.assertEqual(agent_by_str._loss_class, agent_by_class._loss_class)


if __name__ == "__main__":
  np.random.seed(SEED)
  torch.manual_seed(SEED)
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for meta-cfr algorithm."""

import functools
from typing import List
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np

from open_spiel.python.examples.meta_cfr.sequential_games.typing import ApplyFn
from open_spiel.python.examples.meta_cfr.sequential_games.typing import InfostateMapping
from open_spiel.python.examples.meta_cfr.sequential_games.typing import InfostateNode
from open_spiel.python.examples.meta_cfr.sequential_games.typing import Params


def get_batched_input(input_list: List[jax.Array],
                      infostate_list: List[InfostateNode],
                      illegal_action_list: List[List[int]], batch_size: int):
  """Returns list of function arguments extended to be consistent with batch size.

  Args:
    input_list: List of DeviceArrays.
    infostate_list: List of information state nodes.
    illegal_action_list: List of List of illegal actions. Each internal list
      contains illegal actions in each information state.
    batch_size: Batch size.

  Returns:
    input_list, infostate_list, and illegal_action_list with a size consistent
    with batch size (the size of returned arrays are multipliers of batch size).
  """
  items_to_sample = batch_size * (int(len(input_list) / batch_size) +
                                  1) - len(input_list)
  idx_sample = np.random.choice(len(input_list), items_to_sample)
  input_zip = np.array(
      list(zip(input_list, infostate_list, illegal_action_list)),
      dtype=object)
  input_lst_sample = input_zip[idx_sample]
  input_sample, infostate_sample, illegal_action_sample = zip(*input_lst_sample)

  input_list.extend(list(input_sample))
  infostate_list.extend(list(infostate_sample))
  illegal_action_list.extend(list(illegal_action_sample))
  return input_list, infostate_list, illegal_action_list


def mask(cfvalues: np.ndarray, infoset: List[InfostateNode], num_actions: int,
         batch_size: int) -> np.ndarray:
  """Returns counterfactual values of legal actions and put 0 for illegal ones.

  Args:
    cfvalues: Numpy array of counterfactual values.
    infoset: List of information states.
    num_actions: Number of possible actions to take.
    batch_size: Batch size.

  Returns:
    Masked counterfactual values. The counterfactual values of legal actions are
    kept as passed to this function and for illegal actions, we consider 0
    counterfactual value.
  """
  legal_actions = [[infoset[i].world_state.state.legal_actions()] *
                   cfvalues.shape[1] for i in range(batch_size)]

  masked_cfvalues = np.zeros(shape=[batch_size, cfvalues.shape[1], num_actions])
  for i in range(cfvalues.shape[0]):
    for j in range(cfvalues.shape[1]):
      np.put(masked_cfvalues[i][j], legal_actions[i][j], cfvalues[i][j])

  return np.stack(masked_cfvalues)


def filter_terminal_infostates(infostates_map: InfostateMapping):
  """Filter out terminal infostate_node values."""
  return {
      infostate_string: infostate_node
      for infostate_string, infostate_node in infostates_map.items()
      if not infostate_node.is_terminal()
  }


def get_network_output(net_apply: ApplyFn, net_params: Params,
                       net_input: np.ndarray, illegal_actions: List[int],
                       key: hk.PRNGSequence) -> jax.Array:
  """Returns policy generated as output of model.

  Args:
    net_apply: Haiku apply function.
    net_params: Haiku network parameters.
    net_input: Input of the model.
    illegal_actions: List of illegal actions we use to mask the model output.
    key: Pseudo random number.

  Returns:
    Policy generated by model. Model output is filtered to mask illegal actions.
  """
  net_output = jax.jit(net_apply)(net_params, key, net_input)

  if illegal_actions:
    net_output = jnp.delete(net_output, np.array(illegal_actions))

  return jax.nn.softmax(net_output)


def get_network_output_batched(
    net_apply: ApplyFn, net_params: Params, net_input: np.ndarray,
    all_illegal_actions: List[List[int]],
    key: hk.PRNGSequence) -> List[jax.Array]:
  """Returns policy of batched input generated as output of model.

  Args:
    net_apply: Haiku apply function.
    net_params: Haiku network parameters.
    net_input: Input of the model.
    all_illegal_actions: Nested list of illegal actions we use to mask the model
      output. Length of outer list is equal to the batch size.
    key: Pseudo random number.

  Returns:
    List of policies generated by model. Model output is filtered to mask
    illegal actions. Length of the returned list is equal to batch size.
  """
  net_output_batched = net_apply(net_params, next(key), net_input)

  batch_policies = []
  for i, illegal_actions in enumerate(all_illegal_actions):
    net_output = net_output_batched[i]
    if illegal_actions:
      net_output = jnp.expand_dims(
          jnp.delete(net_output, jnp.array(illegal_actions)), axis=0)

    batch_policies.append(jax.nn.softmax(net_output))
  return batch_policies


@functools.partial(jax.jit, static_argnums=(2, 3, 4, 5, 7, 9))
def meta_loss(net_params: Params, cfvalues: np.ndarray,
              net_apply: ApplyFn, steps: int, num_all_actions: int,
              infosets: List[InfostateNode],
              infostate_map: InfostateMapping,
              batch_size: int,
              key: hk.PRNGSequence,
              use_infostate_representation: bool = True) -> float:
  """Meta learning loss function.

  Args:
    net_params: Network parameters.
    cfvalues: Counterfactual values.
    net_apply: Haiku apply function.
    steps: Number of unrolling steps.
    num_all_actions: Number of actions.
    infosets: List of information states.
    infostate_map: Mapping from information state string to information state
      node.
    batch_size: Batch size.
    key: Pseudo random number.
    use_infostate_representation: Boolean value indicating if information state
      representation is used as part of input.

  Returns:
    Mean meta learning loss value.
  """
  regret_sum = np.zeros(shape=[batch_size, 1, num_all_actions])
  total_loss = 0
  step = 0
  infostate_str_one_hot = jnp.expand_dims(
      jnp.array([
          jax.nn.one_hot(infostate_map[infoset.infostate_string],
                         len(infostate_map)) for infoset in infosets
      ]),
      axis=1)

  def scan_body(carry, x):
    del x  # Unused
    regret_sum, current_step, total_loss = carry
    average_regret = regret_sum / (current_step + 1)

    if use_infostate_representation:
      net_input = jnp.concatenate((average_regret, infostate_str_one_hot),
                                  axis=-1)
    else:
      net_input = average_regret
    next_step_x = jax.jit(net_apply)(net_params, key, net_input)
    strategy = jax.nn.softmax(next_step_x)

    value = jnp.matmul(
        jnp.array(cfvalues), jnp.transpose(strategy, axes=[0, 2, 1]))
    curren_regret = jnp.array(cfvalues) - value
    regret_sum += jnp.expand_dims(jnp.mean(curren_regret, axis=1), axis=1)
    current_loss = jnp.mean(
        jnp.max(
            jax.numpy.concatenate(
                [regret_sum,
                 jnp.zeros(shape=[batch_size, 1, 1])],
                axis=-1),
            axis=-1))
    total_loss += current_loss
    current_step += 1
    return (regret_sum, current_step, total_loss), None

  (regret_sum, step, total_loss), _ = jax.lax.scan(
      scan_body, (regret_sum, step, total_loss), None, length=steps)
  return total_loss
